from . import convert  # for export
from . import plots  # for export...
from .regions import GenomicRegions, region_registry
from .regions_from import (
    GenomicRegions_BinnedGenome,
    GenomicRegions_Common,
    GenomicRegions_CommonInAtLeastX,
    GenomicRegions_Difference,
    GenomicRegions_Intersection,
    GenomicRegions_FromBed,
    GenomicRegions_FromBigBed,
    GenomicRegions_FromGFF,
    GenomicRegions_FromPartec,
    GenomicRegions_FromTable,
    GenomicRegions_FromWig,
    GenomicRegions_Invert,
    GenomicRegions_Overlapping,
    GenomicRegions_Union,
    GenomicRegions_Windows,
    GenomicRegions_FilterRemoveOverlapping,
    GenomicRegions_FilterToOverlapping,
)
# from . import annotators


__all__ = [
    "convert",
    "GenomicRegions",
    "GenomicRegions_BinnedGenome",
    "GenomicRegions_Common",
    "GenomicRegions_CommonInAtLeastX",
    "GenomicRegions_Difference",
    "GenomicRegions_FromBed",
    "GenomicRegions_FromBigBed",
    "GenomicRegions_FromGFF",
    "GenomicRegions_FromPartec",
    "GenomicRegions_FromTable",
    "GenomicRegions_FromWig",
    "GenomicRegions_Invert",
    "GenomicRegions_Overlapping",
    "GenomicRegions_Union",
    "GenomicRegions_Intersection",
    "GenomicRegions_Windows",
    "plots",
    "region_registry",
    "GenomicRegions_FilterRemoveOverlapping",
    "GenomicRegions_FilterToOverlapping",
]



from ..regions import GenomicRegions

import numpy as np
import pypipegraph as ppg
import pandas as pd
from pathlib import Path
from mbf_externals.util import lazy_method

_exon_regions_overlapping_cache = {}
_genes_per_genome_singletons = {}


class Genes(GenomicRegions):
    def __new__(cls, genome, alternative_load_func=None, *args, **kwargs):
        """Make sure that Genes for a full genome (ie. before filtering) are singletonic. Ie.
        you can always safely call Genes(my_genome), without worrying about duplicate objects"""
        if alternative_load_func is None:
            if ppg.util.global_pipegraph:
                if not hasattr(
                    ppg.util.global_pipegraph, "_genes_per_genome_singletons"
                ):
                    ppg.util.global_pipegraph._genes_per_genome_singletons = {}
                singleton_dict = ppg.util.global_pipegraph._genes_per_genome_singletons
            else:
                singleton_dict = _genes_per_genome_singletons
            if not genome in singleton_dict:
                singleton_dict[genome] = GenomicRegions.__new__(cls)
            return singleton_dict[genome]
        else:
            return GenomicRegions.__new__(cls)

    def __init__(
        self,
        genome,
        alternative_load_func=None,
        name=None,
        dependencies=None,
        result_dir=None,
        sheet_name=None,
        vid=None,
    ):
        if hasattr(self, "_already_inited"):
            if (
                alternative_load_func is not None
                and alternative_load_func != self.genes_load_func
            ):  # pragma: no cover -
                # this can only happen iff somebody starts to singletonize the Genes
                # with loading_functions - otherwise the duplicate object
                # checker will kick in first
                raise ValueError(
                    "Trying to define Genes(%s) twice with different loading funcs"
                    % self.name
                )
            pass
        else:
            if name is None:
                if alternative_load_func:
                    raise ValueError(
                        "If you pass in an alternative_load_func you also need to specify a name"
                    )
                name = "Genes_%s" % genome.name
                self.top_level = True
            else:
                self.top_level = False
            if alternative_load_func is None:
                load_func = lambda: genome.df_genes.reset_index()  # noqa: E731
            else:
                load_func = alternative_load_func

            self.genes_load_func = load_func
            if result_dir:
                pass
            elif sheet_name:
                result_dir = result_dir or Path("results") / "Genes" / sheet_name / name
            else:
                result_dir = result_dir or Path("results") / "Genes" / name
            result_dir = Path(result_dir).absolute()

            self.column_properties = {
                "chr": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "On which chromosome (or contig) the gene is loacted",
                },
                "start": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Left most position of this gene",
                },
                "stop": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Right most position of this gene",
                },
                "tss": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Position of Transcription start site on the chromosome",
                },
                "tes": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Position of the annotated end of transcription",
                },
                "gene_stable_id": {
                    "user_visible": True,
                    "priority": -1000,
                    "index": True,
                    "description": "Unique identification name for this gene",
                },
                "name": {
                    "user_visible": True,
                    "priority": -999,
                    "index": True,
                    "description": "Offical name for this gene",
                    "nocase": True,
                },
                "strand": {
                    "user_visible": True,
                    "priority": -998,
                    "description": "Which is the coding strand (1 forward, -1 reverse)",
                },
                "description": {
                    "user_visible": True,
                    "priority": -998.5,
                    "description": "A short description of the gene",
                },
            }
            GenomicRegions.__init__(
                self,
                name,
                self._load,
                dependencies if dependencies is not None else [],
                genome,
                on_overlap="ignore",
                result_dir=result_dir,
                sheet_name=sheet_name,
                summit_annotator=False,
            )
            if self.load_strategy.build_deps:
                deps = [genome.download_genome()]
                # function invariant for _load is done by GR.__init__
                deps.append(genome.job_genes())
                deps.append(genome.job_transcripts())
                self.load_strategy.load().depends_on(deps)
            self._already_inited = True
            self.vid = vid

    def __str__(self):
        return "Genes(%s)" % self.name

    def __repr__(self):
        return "Genes(%s)" % self.name

    def register(self):
        pass

    def _load(self):
        """Load func
        """
        if hasattr(self, "df"):  # pragma: no cover
            return
        df = self.genes_load_func()

        if not isinstance(df, pd.DataFrame):
            raise ValueError(
                "GenomicRegion.loading_function must return a DataFrame, was: %s"
                % type(df)
            )
        for col in self.get_default_columns():
            if not col in df.columns:
                func_filename = self.genes_load_func.__code__.co_filename
                func_line_no = self.genes_load_func.__code__.co_firstlineno
                raise ValueError(
                    "%s not in dataframe returned by GenomicRegion.loading_function %s %s - it did return %s"
                    % (col, func_filename, func_line_no, df.columns)
                )
        allowed_chromosomes = set(self.genome.get_chromosome_lengths().keys())
        if len(df):
            for chr in df["chr"]:
                if not chr in allowed_chromosomes:
                    raise ValueError(
                        "Invalid chromosome found when loading %s: '%s', expected one of: %s\nLoading func was %s"
                        % (
                            self.name,
                            chr,
                            sorted(allowed_chromosomes),
                            self.genes_load_func,
                        )
                    )
            if not np.issubdtype(df["tss"].dtype, np.integer):
                raise ValueError(
                    "tss needs to be an integer, was: %s" % df["tss"].dtype
                )
            if not np.issubdtype(df["tes"].dtype, np.integer):
                raise ValueError(
                    "tes needs to be an integer, was: %s" % df["tes"].dtype
                )
            # df = self.handle_overlap(df) Genes don't care about overlap
            if not "start" in df.columns:
                df = df.assign(start=np.min([df["tss"], df["tes"]], 0))
            if not "stop" in df.columns:
                df = df.assign(stop=np.max([df["tss"], df["tes"]], 0))
        else:
            df = df.assign(
                start=np.array([], dtype=np.int32), stop=np.array([], dtype=np.int32)
            )

        if (df["start"] > df["stop"]).any():
            raise ValueError(
                "Genes.loading_function returned a negative interval:\n %s"
                % df[df["start"] > df["stop"]].head()
            )
        self.df = df.sort_values(["chr", "start"], ascending=[True, True]).reset_index(
            drop=True
        )  # since we don't call handle_overlap
        # enforce column order
        cols = ["gene_stable_id", "chr", "start", "stop", "strand", "tss", "tes"]
        for x in df.columns:
            if not x in cols:
                cols.append(x)
        df = df[cols]
        return df

    def get_default_columns(self):
        return ("chr", "tss", "tes", "gene_stable_id")

    def _new_for_filtering(self, new_name, load_func, dependencies, **kwargs):
        """When filtering, a new object of this class is created.
        To pass it the right options from the parent, overwrite this
        """
        return Genes(self.genome, load_func, new_name, dependencies, **kwargs)

    def overlap_genes(self, other_genes):
        if self.genome != other_genes.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_genes.genome)
            )
        if not "gene_stable_id" in other_genes.df.columns:
            raise ValueError(
                "other_genes must be a Genes (or at least have df[:, 'gene_stable_id']"
            )
        return len(
            set(self.df["gene_stable_id"]).intersection(
                other_genes.df["gene_stable_id"]
            )
        )

    @lazy_method
    def regions_tss(self):
        """Return 'point' regions for the transcription start sites, one per gene"""

        def load():
            res = []
            for dummy_idx, row in self.df.iterrows():
                res.append(
                    {
                        "chr": row["chr"],
                        "start": row["tss"],
                        "stop": row["tss"] + 1,
                        "gene_stable_id": row["gene_stable_id"],
                        "tss_direction": row["strand"],
                    }
                )
            return pd.DataFrame(res)

        return GenomicRegions(
            self.name + " TSS", load, [self.load()], self.genome, on_overlap="merge"
        )

    @lazy_method
    def regions_tes(self):
        """Return 'point' regions for the transcription end sites, one per gene"""

        def load():
            res = []
            for dummy_idx, row in self.df.iterrows():
                res.append(
                    {
                        "chr": row["chr"],
                        "start": row["tes"],
                        "stop": row["tes"] + 1,
                        "gene_stable_id": row["gene_stable_id"],
                    }
                )
            return pd.DataFrame(res)

        return GenomicRegions(
            self.name + " TES", load, [self.load()], self.genome, on_overlap="merge"
        )

    @lazy_method
    def regions_introns(self):
        """Return positions of all intronic regions - possibly overlapping"""

        def load():
            res = {
                "chr": [],
                "start": [],
                "stop": [],
                "transcript_stable_id": [],
                "intron_rank": [],
                "gene_stable_id": [],
            }
            canonical_chromosomes = self.genome.get_chromosome_lengths()
            for (transcript_stable_id, transcript_row) in self.genome.df_transcripts[
                ["gene_stable_id", "chr", "strand"]
            ].iterrows():
                if (
                    not transcript_row["chr"] in canonical_chromosomes
                ):  # pragma: no cover
                    continue
                tr = self.genome.transcripts[transcript_stable_id]
                introns = tr.introns
                cardinality = 0
                for start, stop in introns:
                    res["chr"].append(transcript_row["chr"])
                    res["start"].append(start)
                    res["stop"].append(stop)
                    res["transcript_stable_id"].append(transcript_stable_id)
                    res["gene_stable_id"].append(transcript_row["gene_stable_id"])
                    if transcript_row["strand"] == 1:
                        res["intron_rank"].append(cardinality)
                    else:
                        res["intron_rank"].append(len(introns) - 1 - cardinality)
                    cardinality += 1
            return pd.DataFrame(res)

        return GenomicRegions(
            self.name + " introns",
            load,
            [self.load()],
            self.genome,
            on_overlap="ignore",
        )

    @lazy_method
    def regions_exons_overlapping(self):
        """Return positions of all exonic regions - possibly overlapping"""

        if self.load_strategy.build_deps:
            deps = [
                self.load(),
                ppg.FunctionInvariant(
                    "GenomicRegions_{self.name}_exons_overlapping_actual_load",
                    type(self.genome).df_exons,
                ),
            ]
        else:
            deps = []
        return GenomicRegions(
            self.name + "_exons_overlapping",
            lambda: self.genome.df_exons,
            deps,
            self.genome,
            on_overlap="ignore",
        )

    @lazy_method
    def regions_exons_merged(self):
        """Return positions of all exonic regions - possibly overlapping"""
        if self.load_strategy.build_deps:
            deps = [
                self.load(),
                ppg.FunctionInvariant(
                    "GenomicRegions_{self.name}_exons_merged_actual_load",
                    type(self.genome).df_exons,
                ),
            ]
        else:
            deps = []

        return GenomicRegions(
            self.name + "_exons_merged",
            lambda: self.genome.df_exons,
            deps,
            self.genome,
            on_overlap="merge",
        )

    def write_bed(self, output_filename=None):
        """Store the intervals of the GenomicRegion in a BED file"""
        from mbf_fileformats.bed import BedEntry, write_bed

        output_filename = self.pathify(output_filename, self.name + ".bed")

        def write(output_filename=output_filename):
            bed_entries = []
            for dummy_idx, row in self.df.iterrows():
                entry = BedEntry(
                    row["chr"], row["start"], row["stop"], name=row["gene_stable_id"]
                )
                bed_entries.append(entry)
            write_bed(output_filename, bed_entries, {}, self.name, include_header=True)

        if self.load_strategy.build_deps:
            deps = [self.load()]
        else:
            deps = []
        return self.load_strategy.generate_file(output_filename, write, deps)

    def write(self, output_filename=None, mangler_function=None):
        if output_filename is None and mangler_function is not None:
            raise ValueError(
                "plain write (without output filename) is always unmangled"
            )
        output_filename = self.pathify(output_filename, self.get_table_filename())
        output_filename.parent.mkdir(exist_ok=True, parents=True)

        def write_mangler_function(df):
            """Establish a decent column order for genes.write"""
            order = [
                "gene_stable_id",
                "name",
                "chr",
                "start",
                "stop",
                "strand",
                "tss",
                "tes",
                "description",
                "biotype",
            ]
            order = [x for x in order if x in df.columns]
            real_order = order + sorted(
                [
                    x
                    for x in df.columns
                    if x not in order
                    and not x.startswith("mysummit")
                    and not x == "parent_row"
                ]
            )
            if "parent_row" in df.columns:
                real_order += ["parent_row"]
            df = df[real_order]
            if mangler_function:
                df = mangler_function(df.copy())
            return df

        return GenomicRegions.write(
            self, output_filename, mangler_function=write_mangler_function
        )



from pathlib import Path
import pandas as pd
import numpy as np
import pypipegraph as ppg

from .annotator import Annotator
from mbf_externals.util import lazy_method
from mbf_genomics.util import (
    parse_a_or_c_to_column,
    parse_a_or_c_to_anno,
    find_annos_from_column,
)


class DelayedDataFrame(object):
    """Base class for  DataFrame + annotators style of classes.

    An annotator is an object that can calculate additional columns for a DataFrame.

    This is a dual object - it can be used in a pipegraph - in which
    case it .annotate() returns jobs, and calculatin is 'smart & lazy'
    or without a global pipegraph in which case loading and += annotator evaluation
    happens immediatly (though it still chases the dep_anno chain defined by the annotators)

    Annotators may be annotated to any child (created by .filter),
    and in the ppg case, they will be evaluated on the top most parent that
    actually needs them.

    """

    def __init__(self, name, loading_function, dependencies=[], result_dir=None):
        # assert_uniqueness_of_object is taking core of by the load_strategy
        self.name = name
        if result_dir:
            self.result_dir = Path(result_dir)
        else:
            self.result_dir = Path("results") / self.__class__.__name__ / self.name
        self.result_dir.mkdir(parents=True, exist_ok=True)
        if isinstance(loading_function, pd.DataFrame):
            # don't you just love lambda variable binding?
            loading_function = (
                lambda loading_function=loading_function: loading_function
            )

        if not ppg.inside_ppg():
            self.load_strategy = Load_Direct(self, loading_function)
        else:
            self.load_strategy = Load_PPG(self, loading_function, dependencies)
        self.column_to_annotators = {}
        self.annotators = {}
        self.parent = None
        self.children = []
        # this prevents writing the same file with two different mangler functions
        # but still allows you to call write() in ppg settings multiple times
        # if different parts need to ensure it's being written out
        self.mangler_dict = {self.get_table_filename(): None}
        self.load()

    # magic
    def __hash__(self):
        return hash("DelayedDataFrame" + self.name)

    def __str__(self):
        return "DelayedDataFrame(%s)" % self.name

    def __repr__(self):
        return "DelayedDataFrame(%s)" % self.name

    def load(self):
        return self.load_strategy.load()

    def __iadd__(self, other):
        """Add and return self"""
        if isinstance(other, Annotator):
            if ppg.inside_ppg():
                if not self.has_annotator(other):
                    self.load_strategy.add_annotator(other)
                elif self.get_annotator(other.get_cache_name()) is not other:
                    raise ValueError(
                        "trying to add different annotators with identical cache_names\n%s\n%s"
                        % (other, self.get_annotator(other.get_cache_name()))
                    )
            else:
                self.load_strategy.add_annotator(other)

            return self
        else:
            return NotImplemented

    def add_annotator(self, anno):
        """Sugar for
        self += anno
        return self.anno_jobs[anno.get_cache_name()]
        """
        self += anno
        if self.load_strategy.build_deps:  # pragma: no branch
            return self.anno_jobs[anno.get_cache_name()]

    def has_annotator(self, anno):
        return anno.get_cache_name() in self.annotators

    def get_annotator(self, cache_name):
        return self.annotators[cache_name]

    @property
    def root(self):
        x = self
        while x.parent is not None:
            x = x.parent
        return x

    def annotate(self):
        """Job: create plotjobs and dummy annotation job.

        This is all fairly convoluted.
        If you require a particular column, don't depend on this, but on the result of add_annotator().
        This is only the master job that jobs that require *all* columns (table dump...) depend on.
        """
        return self.load_strategy.annotate()

    def filter(
        self,
        new_name,
        df_filter_function,
        annotators=[],
        dependencies=None,
        column_lookup=None,
        **kwargs,
    ):
        """Filter an ddf to a new one called new_name.

        Paramaters
        -----------
            df_filter_function: function|list_of_filters
              function: take a df, return a valid index
              list_of_filters: see DelayedDataFrame.definition_to_function

            annotators:
                annotators used by your filter function.
                Leave empty in the case of list_of_filters

            dependencies:
                list of ppg.Jobs

            column_lookup: offer abbreviations to the column definitions in list_of_filters

        """

        def load():
            idx_or_bools = df_filter_function(self.df)
            if isinstance(idx_or_bools, pd.Index):
                res = self.df.loc[idx_or_bools][self.non_annotator_columns]
                res = res.assign(parent_row=idx_or_bools)
            else:
                res = self.df[idx_or_bools][self.non_annotator_columns]
                res = res.assign(parent_row=self.df.index[idx_or_bools])
            return res

        if dependencies is None:
            dependencies = []
        elif not isinstance(dependencies, list):  # pragma: no cover
            dependencies = list(dependencies)
        if isinstance(df_filter_function, (list, tuple)):
            if isinstance(df_filter_function, (tuple)):
                df_filter_function = [df_filter_function]
            df_filter_function, annotators = self.definition_to_function(
                df_filter_function, column_lookup if column_lookup is not None else {}
            )

        else:
            if isinstance(annotators, Annotator):
                annotators = [annotators]
        if self.load_strategy.build_deps:
            dependencies.append(
                ppg.ParameterInvariant(
                    self.__class__.__name__ + "_" + new_name + "_parent", self.name
                )
            )
            dependencies.append(
                ppg.FunctionInvariant(
                    self.__class__.__name__ + "_" + new_name + "_filter",
                    df_filter_function,
                )
            )
            if hasattr(df_filter_function, "_dependency_params"):
                dependencies.append(
                    ppg.ParameterInvariant(
                        self.__class__.__name__ + "_" + new_name + "_filter",
                        df_filter_function._dependency_params,
                    )
                )

            dependencies.append(self.load())
            for anno in annotators:
                self += anno
                dependencies.append(self.anno_jobs[anno.get_cache_name()])

        else:
            for anno in annotators:
                self += anno

        result = self._new_for_filtering(new_name, load, dependencies, **kwargs)
        result.parent = self
        result.filter_annos = annotators
        for anno in self.annotators.values():
            result += anno

        self.children.append(result)
        return result

    def definition_to_function(self, definition, column_lookup):  # noqa: C901
        """
        Create a filter function from a tuple of
        (column_definition, operator, threshold)

        Operators are strings '>', '<', '==', '>=', '<=', 'isin'
        They may be prefixed by '|' which means 'take absolute first'

        Example:
        genes.filter('2x', [
            ('FDR', '<=', 0.05) # a name from column_lookup
            ('log2FC', '|>', 1),  # absolute by prefixing operator
            ...
            (anno, '>=', 50),
            ((anno, 1), '>=', 50),  # for the second column of the annotator
            ((anno, 'columnX'), '>=', 50),  # for the second column of the annotator
            ('annotator_columnX', '=>' 50), # search for an annotator with that column. Use if exactly one, complain otherwise


        returns: a df_filter_func, [annotators]

        """
        functors = []
        annotators = []
        for column_name, op, threshold in definition:
            anno = None
            if hasattr(self, "df") and column_name in self.df.columns:
                # we can't check for non-annotator columns on filter
                # definition in ppg ddfs that have not been loaded yet.
                # oh well, no worse than passing in a function with an
                # invalid column name
                # exception in the 'we have a df and column is not in it is
                # below
                anno = None
            else:
                if column_name in column_lookup:
                    column_name = column_lookup[column_name]
                try:
                    anno = parse_a_or_c_to_anno(column_name)
                    column_name = parse_a_or_c_to_column(column_name)
                except (ValueError, KeyError):
                    anno = None
                if anno is None:
                    try:
                        annos = find_annos_from_column(column_name)
                    except KeyError:
                        annos = []
                    if len(annos) == 1:
                        anno = annos[0]
                    elif len(annos) > 1:
                        raise KeyError(
                            "Column (%s) was present in multiple annotators: %s.\n Pass in anno or (anno, column)"
                            % (column_name, annos)
                        )

                if anno is None:
                    if hasattr(self, "df"):
                        if not column_name in self.df.columns:
                            raise KeyError(
                                f"unknown column {column_name}",
                                "available",
                                sorted(
                                    set(list(self.df.columns) + list(column_lookup))
                                ),
                            )
                    else:  # I guess, then the filter job fails later.
                        pass
            if op == "==":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ]
                    == threshold
                )  # noqa: E03
            elif op == ">":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ]
                    > threshold
                )  # noqa: E03
            elif op == "<":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ]
                    < threshold
                )  # noqa: E03
            elif op == ">=":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ]
                    >= threshold
                )  # noqa: E03
            elif op == "<=":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ]
                    <= threshold
                )  # noqa: E03
            elif op == "|>":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ].abs()
                    > threshold  # noqa: E03
                )
            elif op == "|<":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ].abs()
                    < threshold
                )  # noqa: E03
            elif op == "|>=":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ].abs()
                    >= threshold
                )  # noqa: E03
            elif op == "|<=":
                f = (
                    lambda df, column_name=column_name, threshold=threshold: df[
                        column_name
                    ].abs()
                    <= threshold
                )  # noqa: E03
            elif op == "isin":
                f = lambda df, column_name=column_name, chosen_set=threshold: df[  # noqa: E731
                    column_name
                ].isin(
                    set(chosen_set)
                )
            else:
                raise ValueError(f"invalid operator {op}")
            functors.append(f)
            if anno is not None:
                annotators.append(anno)

        def filter_func(df):
            keep = np.ones(len(df), bool)
            for f in functors:
                keep &= f(df)
            return keep
        filter_func._dependency_params = (definition, column_lookup)
        return filter_func, annotators

    def _new_for_filtering(self, new_name, load_func, deps, **kwargs):
        if not "result_dir" in kwargs:
            kwargs["result_dir"] = self.result_dir / new_name
        return self.__class__(new_name, load_func, deps, **kwargs)

    def clone_without_annotators(
        self, new_name, non_annotator_columns_to_copy=None, result_dir=None
    ):
        """Create a clone of this DelayedDataFrame without any of the old annotators attached"""
        if isinstance(non_annotator_columns_to_copy, str):
            raise ValueError(
                "non_annotator_columns_to_copy must be an iterable - maybe you were trying to set result_dir?"
            )

        def load():
            return self.df[self.non_annotator_columns]

        deps = [self.load()]
        result = self.__class__(new_name, load, deps, result_dir)
        return result

    def get_table_filename(self):
        return self.result_dir / (self.name + ".tsv")

    def mangle_df_for_write(self, df):
        return df

    def write(self, output_filename=None, mangler_function=None, float_format="%4g"):
        """Job: Store the internal DataFrame (df) in a table.
        To sort, filter, remove columns, etc before output,
        pass in a mangler_function (takes df, returns df)

        Retruns a (Job, Path) tuple - job is None if outside ppg
        """
        output_filename = self.pathify(
            output_filename, self.get_table_filename().absolute()
        )

        def write(output_filename):
            if mangler_function:
                df = mangler_function(self.df.copy())
            else:
                df = self.mangle_df_for_write(self.df)
            if str(output_filename).endswith(".xls") or str(output_filename).endswith(
                ".xlsx"
            ):
                try:
                    df.to_excel(output_filename, index=False, float_format=float_format)
                except (ValueError):
                    df.to_csv(
                        output_filename,
                        sep="\t",
                        index=False,
                        float_format=float_format,
                    )
            else:
                df.to_csv(
                    output_filename,
                    sep="\t",
                    index=False,
                    encoding="utf-8",
                    float_format=float_format,
                )

        if self.load_strategy.build_deps:
            deps = [
                self.annotate(),
                ppg.FunctionInvariant(
                    str(output_filename) + "_mangler", mangler_function
                ),
                ppg.ParameterInvariant(str(output_filename), float_format),
            ]
        else:
            deps = []
        return self.load_strategy.generate_file(output_filename, write, deps)

    def plot(self, output_filename, plot_func, calc_func=None, annotators=None):
        output_filename = self.pathify(output_filename)

        def do_plot(output_filename=output_filename):
            df = self.df
            if calc_func is not None:
                df = calc_func(df)
            p = plot_func(df)
            if hasattr(p, "pd"):
                p = p.pd
            p.save(output_filename, verbose=False)

        if self.load_strategy.build_deps:
            deps = [
                ppg.FunctionInvariant(
                    output_filename.with_name(output_filename.name + "_plot_func"),
                    plot_func,
                )
            ]
            if annotators is None:
                deps.append(self.annotate())
            else:
                deps.extend([self.add_annotator(x) for x in annotators])
        else:
            deps = []
            if annotators is not None:
                for anno in annotators:
                    self += anno
        return self.load_strategy.generate_file(output_filename, do_plot, deps)

    def pathify(self, output_filename, default=None):
        """Turn output_filename into a Path. If it's a relative path, treat
        it as relative to self.result_dir,
        if it's absolute, take it is at is
        """
        if output_filename is None:
            output_filename = default
        output_filename = Path(output_filename)
        if not output_filename.is_absolute():
            output_filename = self.result_dir / output_filename
        return output_filename.absolute()


def _combine_annotator_df_and_old_df(a_df, ddf_df):
    if len(a_df) == len(ddf_df):
        if isinstance(a_df.index, pd.RangeIndex) and a_df.index.start == 0:
            # assume it is in order
            a_df.index = ddf_df.index
    else:
        raise ValueError(
            "Length and index mismatch - annotator did not return enough rows"
        )
    new_df = pd.concat([ddf_df, a_df], axis=1)
    if len(new_df) != len(ddf_df):
        raise ValueError(
            "Index mismatch between DataFrame and Annotator result concating added %i rows- "
            % (len(new_df) - len(ddf_df))
            + "Annotator must return either a DF with a compatible index "
            "or one with a RangeIndex(0,len(df))"
        )
    return new_df


class Load_Direct:
    def __init__(self, ddf, loading_function):
        self.ddf = ddf
        self.loading_function = loading_function
        self.build_deps = False

    def load(self):
        if not hasattr(self.ddf, "df"):
            self.ddf.df = self.loading_function()
            self.ddf.non_annotator_columns = self.ddf.df.columns

    def generate_file(self, filename, write_callback, dependencies, empty_ok=False):
        write_callback(filename)
        return None, Path(filename)

    def add_annotator(self, anno):
        if anno.get_cache_name() in self.ddf.annotators:
            if self.ddf.annotators[anno.get_cache_name()] != anno:
                raise ValueError(
                    "Trying to add two different annotators with same cache name: %s and %s"
                    % (anno, self.ddf.annotators[anno.get_cache_name()])
                )
            return
        self.ddf.annotators[anno.get_cache_name()] = anno
        for d in anno.dep_annos():
            if d is not None:
                self.ddf += d
        s_should = set(anno.columns)
        if not s_should:
            raise IndexError("Empty columns")
        if (
            self.ddf.parent is not None
            and anno.get_cache_name() in self.ddf.parent.annotators
        ):
            a_df = self.ddf.parent.df.loc[self.ddf.df.index]
            a_df = a_df[s_should]
        else:
            if not isinstance(anno.columns, list):
                raise ValueError("Columns was not a list")
            if hasattr(anno, "calc_ddf"):
                a_df = anno.calc_ddf(self.ddf)
            else:
                a_df = anno.calc(self.ddf.df)
        if isinstance(a_df, pd.Series) and len(s_should) == 1:
            a_df = pd.DataFrame({next(iter(s_should)): a_df})
        elif not isinstance(a_df, pd.DataFrame):
            raise ValueError(
                "Annotator return non DataFrame result (nor a Series and len(anno.columns) == 1)"
            )
        s_actual = set(a_df.columns)
        if s_should != s_actual:
            raise ValueError(
                "Annotator declared different columns from those actualy calculated: %s"
                % (s_should.symmetric_difference(s_actual))
            )
        for k in s_actual:
            if k in self.ddf.df.columns:
                raise ValueError("Same column form two annotators", k)
        self.ddf.df = _combine_annotator_df_and_old_df(a_df, self.ddf.df)

        for c in self.ddf.children:
            c += anno

    def annotate(self):  # a noop
        return None


class Load_PPG:
    def __init__(self, ddf, loading_function, deps):
        ppg.assert_uniqueness_of_object(ddf)
        ddf.cache_dir = (
            Path(ppg.util.global_pipegraph.cache_folder)
            / ddf.__class__.__name__
            / ddf.name
        )
        ddf.cache_dir.mkdir(parents=True, exist_ok=True)
        self.ddf = ddf
        self.ddf.anno_jobs = {}

        self.loading_function = loading_function
        self.deps = deps
        self.build_deps = True
        self.tree_fixed = False

    def add_annotator(self, anno):
        if ppg.util.global_pipegraph.running:
            raise ppg.JobContractError(
                "Can not add_annotator in a running pipegraph"
                " - the annotator structure get's fixed when a "
                "pipegraph is run, you can't add to it in e.g. a "
                "JobGeneratingJob"
            )
        cache_name = anno.get_cache_name()
        forbidden_chars = "/", "?", "*"
        if any((x in cache_name for x in forbidden_chars)) or len(cache_name) > 60:
            raise ValueError(
                "annotator.column_names[0] not suitable as a cache_name (was %s), add cache_name property"
                % repr(cache_name)
            )
        if not cache_name in self.ddf.annotators:
            # if not hasattr(anno, "columns"): # handled by get_cache_name
            # raise AttributeError("no columns property on annotator %s" % repr(anno))
            self.ddf.annotators[cache_name] = anno
            self.ddf.anno_jobs[cache_name] = self.get_anno_dependency_callback(anno)
        for c in self.ddf.children:
            c += anno
        return self.ddf.anno_jobs[cache_name]

    @lazy_method
    def load(self):
        def load_func(df):
            self.ddf.df = df
            self.ddf.non_annotator_columns = self.ddf.df.columns

        job = ppg.CachedDataLoadingJob(
            self.ddf.cache_dir / "calc", self.loading_function, load_func
        )
        job.depends_on(self.deps).depends_on(
            ppg.FunctionInvariant(
                self.ddf.__class__.__name__ + "_" + self.ddf.name + "_load",
                self.loading_function,
            )
        )
        return job

    def generate_file(self, filename, write_callback, dependencies, empty_ok=False):
        return (
            ppg.FileGeneratingJob(filename, write_callback, empty_ok=empty_ok)
            .depends_on(dependencies)
            .depends_on(self.load()),
            Path(filename),
        )

    def get_anno_dependency_callback(self, anno):
        if anno.get_cache_name() in self.ddf.anno_jobs:
            raise NotImplementedError("Should have checked before")  # pragma: no cover

        def gen():
            self.ddf.root.load_strategy.fix_anno_tree()
            return self.ddf.anno_jobs[anno.get_cache_name()]

        return gen

    def fix_anno_tree(self):
        if self.ddf.parent is not None:  # pragma: no branch
            raise NotImplementedError(
                "Should only be called on the root"
            )  # pragma: no cover
        if self.tree_fixed:
            return
        self.tree_fixed = True

        def recursivly_add_annos(deps, a):
            new = a.dep_annos()
            for n in new:
                if n is None:
                    continue
                if n not in deps:
                    recursivly_add_annos(deps, n)
                    deps.add(n)

        def descend_and_add_annos(node):
            annos_here = set(node.ddf.annotators.values())
            deps = set()
            for a in annos_here:
                recursivly_add_annos(deps, a)
            annos_here.update(deps)
            for anno in annos_here:
                node.add_annotator(anno)
                for c in node.ddf.children:
                    c.load_strategy.add_annotator(anno)
            for (
                c
            ) in (
                node.ddf.children
            ):  # they might have annos that the parent did not have
                descend_and_add_annos(c.load_strategy)

        def descend_and_jobify(node):
            for anno in node.ddf.annotators.values():
                if (
                    node.ddf.parent is not None
                    and anno.get_cache_name() in node.ddf.parent.annotators
                ):
                    node.ddf.anno_jobs[anno.get_cache_name()] = node._anno_load(anno)
                else:
                    # this is the top most node with this annotator
                    node.ddf.anno_jobs[
                        anno.get_cache_name()
                    ] = node._anno_cache_and_calc(anno)
                if hasattr(anno, "register_qc"):
                    anno.register_qc(node.ddf)
            for c in node.ddf.children:
                descend_and_jobify(c.load_strategy)

        descend_and_add_annos(self)
        for anno in self.ddf.annotators.values():
            job = self._anno_cache_and_calc(anno)
            self.ddf.anno_jobs[anno.get_cache_name()] = job
        # for c in self.ddf.children:
        descend_and_jobify(self)

    def _anno_load(self, anno):
        def load():
            self.ddf.df = pd.concat(
                [
                    self.ddf.df,
                    self.ddf.parent.df[anno.columns].reindex(self.ddf.df.index),
                ],
                axis=1,
            )

        job = ppg.DataLoadingJob(self.ddf.cache_dir / anno.get_cache_name(), load)
        job.depends_on(
            ppg.FunctionInvariant(
                self.ddf.cache_dir / (anno.get_cache_name() + "_funcv"), anno.calc
            ),
            self.ddf.parent.anno_jobs[anno.get_cache_name()],
            self.ddf.load(),
        )
        return job

    def _anno_cache_and_calc(self, anno):
        def calc():
            if not isinstance(anno.columns, list):
                raise ValueError("Columns was not a list")

            if hasattr(anno, "calc_ddf"):
                df = anno.calc_ddf(self.ddf)
            else:
                df = anno.calc(self.ddf.df)
            if isinstance(df, pd.Series) and len(anno.columns) == 1:
                df = pd.DataFrame({anno.columns[0]: df})
            if not isinstance(df, pd.DataFrame):
                raise ValueError(
                    "result was no dataframe (or series and len(anno.columns) == 1)"
                )
            return df

        def load(df):
            s_should = set(anno.columns)
            if not len(s_should):
                raise ValueError("anno.columns was empty")
            s_actual = set(df.columns)
            if s_should != s_actual:
                raise ValueError(
                    "Annotator declared different columns from those actualy calculated: %s"
                    % (s_should.symmetric_difference(s_actual))
                )
            if set(df.columns).intersection(self.ddf.df.columns):
                raise ValueError(
                    "Annotator created columns that were already present.",
                    self.ddf.name,
                    anno.get_cache_name(),
                    set(df.columns).intersection(self.ddf.df.columns),
                )
            self.ddf.df = _combine_annotator_df_and_old_df(df, self.ddf.df)

        (self.ddf.cache_dir / anno.__class__.__name__).mkdir(exist_ok=True)
        job = ppg.CachedDataLoadingJob(
            self.ddf.cache_dir / anno.__class__.__name__ / anno.get_cache_name(),
            calc,
            load,
        )
        ppg.Job.depends_on(
            job, self.load()
        )  # both the load and nthe calc needs our ddf.df
        job.depends_on(
            self.load(),
            ppg.FunctionInvariant(
                self.ddf.cache_dir / (anno.get_cache_name() + "_calc_func"),
                anno.calc if hasattr(anno, "calc") else anno.calc_ddf,
            ),
        )
        for d in anno.dep_annos():
            if d is not None:
                job.depends_on(self.ddf.anno_jobs[d.get_cache_name()])
        job.depends_on(anno.deps(self.ddf))
        job.lfg.cores_needed = getattr(anno, "cores_needed", 1)
        return job

    def annotate(self):
        res = lambda: list(self.ddf.anno_jobs.values())  # noqa: E731
        res.job_id = self.ddf.name + "_annotate_callback"
        return res



import pandas as pd


def read_pandas(filename):
    filename = str(filename)
    if filename.endswith(".xls") or filename.endswith(".xlsx"):
        from xlrd import XLRDError

        try:
            filein = pd.read_excel(filename)
        except XLRDError:
            filein = pd.read_csv(filename, sep="\t")
        return filein

    elif filename.endswith(".tsv"):
        return pd.read_csv(filename, sep="\t")
    elif filename.endswith(".csv"):
        return pd.read_csv(filename)
    else:
        raise ValueError("Unknown filetype: %s" % filename)


def freeze(obj):
    """ Turn dicts into frozendict,
        lists into tuples, and sets
        into frozensets, recursively - usefull
        to get a hash value..
    """
    # TODO: combine with ppg.util.freeze
    import pypipegraph as ppg
    return ppg.util.freeze(obj)


def parse_a_or_c(ac):
    """parse  an annotator/column combo into a tuple
    anno, column

    Input may be
            a str -> None, column name
            an annotator -> anno, anno.columns[0]
            an (annotator, str) tuple -> anno, str
            an (annotator, int(i)) tuple -> anno, annotator.columns[i]
    """
    from mbf_genomics.annotator import Annotator

    if isinstance(ac, str):
        return (None, ac)
    elif isinstance(ac, Annotator):
        return ac, ac.columns[0]
    elif isinstance(ac, tuple) and len(ac) == 2 and isinstance(ac[0], Annotator):
        if isinstance(ac[1], int):
            return ac[0], ac[0].columns[ac[1]]
        else:
            if not ac[1] in ac[0].columns:
                raise KeyError(
                    "Invalid column name, %s -annotator had %s", (ac[1], ac[0].columns)
                )
            return ac
    elif isinstance(ac, tuple) and len(ac) == 2 and ac[0] is None:
        return ac
    else:
        raise ValueError("parse_a_or_c could not parse %s" % (ac,))


def parse_a_or_c_to_column(k):
    """Parse an annotator + column spec to the column name.
    See parse_a_or_c
    """
    return parse_a_or_c(k)[1]


def parse_a_or_c_to_anno(k):
    """Parse an annotator + column spec to the annotator (or None)
    See parse_a_or_c
    """
    return parse_a_or_c(k)[0]


def parse_a_or_c_to_plot_name(k, default=None):
    """Parse an annotator + column spec to a plot name
    See parse_a_or_c_to_column

    Defaults to column name if no plot_name is defined on annotator

    """
    ac = parse_a_or_c(k)
    if ac[0] is None:
        return k
    return getattr(ac[0], "plot_name", default if default is not None else ac[1])


def find_annos_from_column(k):
    from . import annotator
    import pypipegraph as ppg

    if ppg.inside_ppg():
        if not hasattr(ppg.util.global_pipegraph, "_annotator_singleton_dict"):
            ppg.util.global_pipegraph._annotator_singleton_dict = {}
        singleton_dict = ppg.util.global_pipegraph._annotator_singleton_dict
    else:
        singleton_dict = annotator.annotator_singletons

    res = []
    for anno in singleton_dict["lookup"]:
        if k in anno.columns:
            res.append(anno)
    if res:
        return res
    else:
        raise KeyError("No anno for column '%s' found" % (k,))



import pytest


@pytest.fixture
def clear_annotators(request):
    """Clear the annotator singleton instance cache
    which is only used if no ppg is in play"""
    import mbf_genomics.annotator

    mbf_genomics.annotator.annotator_singletons.clear()
    mbf_genomics.annotator.annotator_singletons["lookup"] = []



from ..annotator import Annotator
import pandas as pd


class SummitBase(Annotator):
    pass


class SummitMiddle(SummitBase):
    """Place a summit right in the center (ie. a fake summit"""

    columns = ["summit middle"]
    column_properties = {
        columns[0]: {
            "description": "Fake summit, just the center of the region (given relative to start)"
        }
    }

    def calc(self, df):
        res = []
        for dummy_idx, row in df.iterrows():
            res.append((row["stop"] + row["start"]) / 2 - row["start"])
        return pd.Series(res)


# from ..genes.anno_tag_counts import GRUnstrandedRust as TagCount
# from ..genes.anno_tag_counts import GRStrandedRust as TagCountStranded
from ..genes.anno_tag_counts import _NormalizationAnno


class NormalizationCPM(_NormalizationAnno):
    """Normalize to 1e6 by taking the sum of all genes"""

    def __init__(self, base_column_spec):
        self.name = "CPM(lane)"
        self.normalize_to = 1e6
        super().__init__(base_column_spec)
        self.column_properties = {
            self.columns[0]: {"description": "Tag count normalized to lane tag count"}
        }

    def calc(self, df):
        raw_counts = df[self.raw_column]
        total = max(
            1, sum((x.mapped for x in self.raw_anno.aligned_lane.get_bam().get_index_statistics()))
        )
        result = raw_counts * (self.normalize_to / total)
        return pd.Series(result)



from mbf_genomics.annotator import Annotator, FromFile
from typing import Dict, List
from pypipegraph import Job
from mbf_genomics import DelayedDataFrame
from pandas import DataFrame
import pandas as pd
import hashlib


class Description(Annotator):
    """Add the description for the genes from genome.

    @genome may be None (default), then the ddf is queried for a '.genome'
    Requires a genome with df_genes_meta - e.g. EnsemblGenomes
    """

    columns = ["description"]

    def __init__(self, genome=None):
        self.genome = genome

    def calc_ddf(self, ddf):
        if self.genome is None:
            try:
                genome = ddf.genome
            except AttributeError:
                raise AttributeError(
                    "ddf had no .genome and no genome was passed to Description"
                )
        else:
            genome = self.genome
        lookup = dict(genome.df_genes_meta["description"].items())
        result = []
        for gene_stable_id in ddf.df["gene_stable_id"]:
            result.append(lookup.get(gene_stable_id, ""))
        return pd.Series(result, index=ddf.df.index)


def GeneStrandedSalmon(*args, **kwargs):
    """Deprecated. use anno_tag_counts.Salmon
    """
    raise NotImplementedError("Deprecated. Use anno_tag_counts.Salmon")


# FromFile forwarded to mbf_genomics.annotator.FromFile
FromFile = FromFile


class TMM(Annotator):
    """
    Calculates the TMM normalization from edgeR on some raw counts.
    
    Parameters
    ----------
    raw : Dict[str, Annotator]
        Dictionary of raw count annotator for all samples.
    dependencies : List[Job], optional
        List of additional dependencies, by default [].
    samples_to_group : Dict[str, str], optional
        A dictionary sample name to group name, by default None.
    """

    def __init__(
        self,
        raw: Dict[str, Annotator],
        dependencies: List[Job] = None,
        samples_to_group: Dict[str, str] = None,
    ):
        """Constructor."""
        import mbf_r # so we fail early

        self.sample_column_lookup = {}
        for sample_name in raw:
            self.sample_column_lookup[
                sample_name
            ] = f"{raw[sample_name].columns[0]} TMM"
        self.columns = list(self.sample_column_lookup.values())
        self.dependencies = []
        if dependencies is not None:
            self.dependencies = dependencies
        self.raw = raw
        self.samples_to_group = samples_to_group
        self.cache_name = hashlib.md5(self.columns[0].encode("utf-8")).hexdigest()

    def calc_ddf(self, ddf: DelayedDataFrame) -> DataFrame:
        """
        Calculates TMM columns to be added to the ddf instance.

        TMM columns are calculated using edgeR with all samples given in self.raw.

        Parameters
        ----------
        ddf : DelayedDataFrame
            The DelayedDataFrame instance to be annotated.

        Returns
        -------
        DataFrame
            A dataframe containing TMM normalized columns for each
        """
        raw_columns = [self.raw[sample_name].columns[0] for sample_name in self.raw]
        df = ddf.df[raw_columns]
        df_res = self.call_edgeR(df)
        rename = {}
        for col in df_res.columns:
            rename[col] = f"{col} TMM"
        df_res = df_res.rename(columns=rename)
        return df_res

    def call_edgeR(self, df_counts: DataFrame) -> DataFrame:
        """
        Call to edgeR via r2py to get TMM (trimmed mean of M-values)
        normalization for raw counts.

        Prepare the edgeR input in python and call edgeR calcNormFactors via
        r2py. The TMM normalized values are returned in a DataFrame which
        is converted back to pandas DataFrame via r2py.

        Parameters
        ----------
        df_counts : DataFrame
            The dataframe containing the raw counts.

        Returns
        -------
        DataFrame
            A dataframe with TMM values (trimmed mean of M-values).
        """
        import mbf_r
        import rpy2.robjects as ro
        # import rpy2.robjects.numpy2ri as numpy2ri

        ro.r("library(edgeR)")
        df_input = df_counts
        to_df = {"lib.size": df_input.sum(axis=0)}
        if self.samples_to_group is not None:
            to_df["group"] = [
                self.samples_to_group[sample_name]
                for sample_name in self.samples_to_group
            ]
        df_samples = pd.DataFrame(to_df)
        r_counts = mbf_r.convert_dataframe_to_r(df_input)
        r_samples = mbf_r.convert_dataframe_to_r(df_samples)
        y = ro.r("DGEList")(counts=r_counts, samples=r_samples,)
        # apply TMM normalization
        y = ro.r("calcNormFactors")(y)
        norm_factors = ro.r("function(y){y$samples}")(y)
        result = mbf_r.convert_dataframe_from_r(norm_factors)
        df = df_counts.divide(result["norm.factors"].values)
        df = df.reset_index(drop=True)
        return df

    def deps(self, ddf) -> List[Job]:
        """Return ppg.jobs"""
        return self.dependencies

    def dep_annos(self) -> List[Annotator]:
        """Return other annotators"""
        return list(self.raw.values())



"""Utility functions to pass to GenomicRegions.convert(..., convert_func)

"""

# TODO: Liftover utility -> mbf_externals
# TODO: chains not included?

import tempfile
import pandas as pd
import pypipegraph as ppg
import numpy as np
import subprocess
from mbf_externals.util import to_string, to_bytes
from pathlib import Path

file_path = Path(__file__).parent


def grow(basepairs):
    """A function for GenomicRegions.convert that enlarges the regions
    bei doing start = start - basepairs, stop = stop + basepairs"""

    def do_grow(df):
        starts = df["start"] - basepairs
        starts[starts < 0] = 0
        stops = df["stop"] + basepairs
        new_df = df.copy()
        new_df = new_df.assign(start=starts, stop=stops)
        return new_df

    return do_grow, [], basepairs


def promotorize(basepairs=1250):
    """Genes.convert - returns [-basepairs...tss] regions"""

    def do_promotorize(df):
        res = {"chr": df["chr"]}
        res["start"] = np.zeros((len(df),), dtype=np.int32)
        res["stop"] = np.zeros((len(df),), dtype=np.int32)
        forward = df["strand"] == 1
        res["start"][:] = df["tss"]  # Assign within array.
        res["stop"][:] = df["tss"]  # Assign within array.
        res["start"][forward] -= basepairs
        res["start"][res["start"] < 0] = 0
        res["stop"][~forward] += basepairs
        res["strand"] = df["strand"]
        res["gene_stable_id"] = df["gene_stable_id"]
        res = pd.DataFrame(res)
        res = res[res["start"] != res["stop"]]  # can happen by the 0 limiting
        return res

    return do_promotorize, [], basepairs


def shift(basepairs):
    def do_shift(df):
        res = {
            "chr": df["chr"],
            "start": df["start"] + basepairs,
            "stop": df["stop"] + basepairs,
        }
        return pd.DataFrame(res)

    return do_shift


def summit(summit_annotator):
    def do_summits(df):
        summit_col = summit_annotator.columns[0]
        starts = (df["start"] + df[summit_col]).astype(np.int)
        res = {"chr": df["chr"], "start": starts, "stop": starts + 1}
        return pd.DataFrame(res)

    return do_summits, [summit_annotator]


def merge_connected():
    """Merge regions that are next to each other.
    100..200, 200..300 becomes 100..300
    """

    def do_merge(df):
        from mbf_nested_intervals import merge_df_intervals

        return merge_df_intervals(df, lambda iv: iv.merge_connected())

    return do_merge


class LiftOver(object):
    def __init__(self):
        import mbf_genomes
        from mbf_externals.kent import LiftOver as LiftOverAlgorithm

        self.data_path = mbf_genomes.data_path / "liftovers"
        self.replacements = {"hg19to38": {"11_gl000202_random": "GL000202.1"}}
        self.algo = LiftOverAlgorithm()
        self.algo.store.unpack_version(self.algo.name, self.algo.version)

    def do_liftover(self, listOfChromosomeIntervals, chain_file):
        """perform a lift over. Error messages are silently swallowed!"""
        tmp_input = tempfile.NamedTemporaryFile(mode="wb")
        tmp_output = tempfile.NamedTemporaryFile(mode="wb")
        tmp_error = tempfile.NamedTemporaryFile(mode="wb")
        max_len = 0
        listOfChromosomeIntervals = [list(row) for row in listOfChromosomeIntervals]
        for row in listOfChromosomeIntervals:
            tmp_input.write(b" ".join(to_bytes(str(x)) for x in row))
            tmp_input.write(b"\n")
            max_len = max(len(row), max_len)
        tmp_input.write(b"\n")
        tmp_input.flush()  # it's magic ;)
        cmd = [
            self.algo.path / "liftOver",
            tmp_input.name,
            chain_file,
            tmp_output.name,
            tmp_error.name,
        ]
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        dummy_stdout, stderr = p.communicate()
        if p.returncode != 0:  # pragma: no cover
            raise ValueError(
                "do_liftover failed. Returncode: %s, stderr: %s"
                % (p.returncode, stderr)
            )
        tmp_output_in = open(tmp_output.name, "rb")
        res = []
        for row in tmp_output_in:
            row = row.strip().split(b"\t")
            row[0] = to_string(row[0])
            row[1] = int(row[1])
            row[2] = int(row[2])
            res.append(tuple(row))
        tmp_error_in = open(tmp_error.name, "rb")
        tmp_error_in.read()
        tmp_input.close()
        tmp_output.close()
        tmp_error.close()
        return res

    def get_convert_func(self, key, keep_name=False, filter_to_these_chromosomes=None):
        """Note that filter_to_these_chromosomes is after the replacements have kicked in"""
        chain_file = self.data_path / (key + ".over.chain")
        if not chain_file.exists():  # pragma: no cover
            raise ValueError("invalid liftover key, file not found: %s" % chain_file)
        if filter_to_these_chromosomes:
            filter_to_these_chromosomes = set(filter_to_these_chromosomes)

        def do_convert(df):
            if df.index.duplicated().any():  # pragma: no cover
                raise ValueError("liftover only works with unique indices")
            df.index = [str(x) for x in df.index]
            input_tuples = [
                ("chr" + row["chr"], row["start"], row["stop"], idx)
                for idx, row in df.iterrows()
            ]

            output_tuples = self.do_liftover(input_tuples, chain_file)
            output_lists = list(zip(*output_tuples))
            res = pd.DataFrame(
                {
                    "chr": output_lists[0],
                    "start": output_lists[1],
                    "stop": output_lists[2],
                    "parent": [x.decode("utf-8") for x in output_lists[3]],
                }
            ).set_index("parent")
            new_chr = []
            for x in res["chr"]:
                x = x[3:]
                # these are untested as of 2019-03-27
                if x == "m":  # pragma: no cover
                    x = "MT"
                elif (
                    key in self.replacements and x in self.replacements[key]
                ):  # pragma: no cover
                    x = self.replacements[key][x]
                new_chr.append(x)
            res["chr"] = new_chr
            for col in df.columns:
                if col not in res.columns:
                    res = res.assign(**{col: df[col]})
            if filter_to_these_chromosomes:
                res = res[res["chr"].isin(filter_to_these_chromosomes)]
            return res

        if ppg.inside_ppg():
            do_convert.dependencies = [
                ppg.FileTimeInvariant(chain_file),
                ppg.FunctionInvariant(
                    "genomics.regions.convert.LiftOver.do_liftover",
                    LiftOver.do_liftover,
                ),
            ]
        return do_convert


def lift_over(from_to, keep_name=False, filter_to_these_chromosomes=None):
    """Map a genome to another genome.
    from_to looks like hg19ToHg38
    see mbf_genomes/data/liftovers for the list currently supported"""
    return LiftOver().get_convert_func(
        from_to,
        keep_name=keep_name,
        filter_to_these_chromosomes=filter_to_these_chromosomes,
    )


def cookie_cutter(bp):
    """ transform all their binding regions to -1/2 * bp ... 1/2 * bp centered
    around the old midpoint... (so pass in the final size of the region)
    inspired by Lupien et al (doi 10.1016/j.cell.2008.01.018")
    """

    def convert(df):
        peak_lengths = df["stop"] - df["start"]
        centers = np.array(df["start"] + peak_lengths // 2, dtype=np.int32)
        new_starts = centers - bp // 2
        new_stops = new_starts + bp
        new_starts[new_starts < 0] = 0
        res = pd.DataFrame({"chr": df["chr"], "start": new_starts, "stop": new_stops})
        if "strand" in df.columns:  # pragma: no branch
            res["strand"] = df["strand"]
        return res

    return convert, [], bp


def cookie_summit(summit_annotator, bp, drop_those_outside_chromosomes=False):
    """ transform all their binding regions to -1/2 * bp ... 1/2 * bp centered
    around the summit (so pass in the final size of the region)

    if @drop_those_outside_chromosomes is set, regions < 0 are dropped
    """

    def do_summits(df):
        summit_col = summit_annotator.columns[0]
        res = {
            "chr": df["chr"],
            "start": df["start"] + df[summit_col].astype(int) - bp // 2,
            "stop": df["start"] + df[summit_col].astype(int) + bp // 2,
        }
        res = pd.DataFrame(res)
        if drop_those_outside_chromosomes:
            res = res[res["start"] >= 0]
        else:
            res = res.assign(start=res["start"].clip(lower=0))
        return res

    return do_summits, [summit_annotator], (bp, drop_those_outside_chromosomes)


def windows(window_size, drop_smaller_windows=False):
    """Chuck the region into window_size sized windows.
    if @drop_smaller_windows is True, the right most windows get chopped"""

    def create_windows(df):
        res = {"chr": [], "start": [], "stop": []}
        for dummy_idx, row in df.iterrows():
            for start in range(row["start"], row["stop"], window_size):
                stop = min(start + window_size, row["stop"])
                if drop_smaller_windows and stop - start < window_size:
                    continue
                res["chr"].append(row["chr"])
                res["start"].append(start)
                res["stop"].append(stop)
        return pd.DataFrame(res)

    return create_windows, [], (window_size, drop_smaller_windows)



import pypipegraph as ppg
import pandas as pd
import numpy as np
import math

from .regions import GenomicRegions
from mbf_nested_intervals import merge_df_intervals
from mbf_externals.util import to_string
from ..util import read_pandas


def verify_same_genome(list_of_grs):
    if len(set([x.genome for x in list_of_grs])) > 1:
        raise ValueError("Mixing GenomicRegions from different genomes not supported")


def GenomicRegions_FromGFF(
    name,
    filename,
    genome,
    filter_function=None,
    comment_char=None,
    on_overlap="raise",
    chromosome_mangler=None,
    fix_negative_coordinates=False,
    alternative_class=None,
    summit_annotator=None,
    vid=None,
):
    """Create a GenomicRegions from a gff file.
    You can filter entries with @filter_function(gff_entry_dict) -> Bool,
    remove comment lines starting with a specific character with @comment_char,
    mangle the chromosomes with @chromosome_mangler(str) -> str,
    replace negative coordinates with 0 (@fix_negative_coordinates),
    or provide an alternative constructor to call with @alternative_class
    """

    def load():
        from mbf_fileformats.gff import gffToDict

        entries = gffToDict(filename, comment_char=comment_char)
        data = {
            "chr": [],
            "start": [],
            "stop": [],
            "score": [],
            "strand": [],
            "name": [],
        }
        name_found = False
        for entry in entries:
            if filter_function and not filter_function(entry):
                continue
            if chromosome_mangler:
                chr = chromosome_mangler(entry["seqname"])
            else:
                chr = entry["seqname"]
            data["chr"].append(to_string(chr))
            start = entry["start"]
            if fix_negative_coordinates and start < 0:
                start = 0
            data["start"].append(start)
            data["stop"].append(entry["end"])
            data["score"].append(entry["score"])
            data["strand"].append(entry["strand"])
            name = entry["attributes"].get("Name", [""])[0]
            data["name"].append(name)
            if name:
                name_found = True
        if not name_found:
            del data["name"]
        return pd.DataFrame(data)

    if alternative_class is None:  # pragma: no cover
        alternative_class = GenomicRegions
    if ppg.inside_ppg():
        deps = [
            ppg.FileTimeInvariant(filename),
            ppg.ParameterInvariant(
                name + "_params_GenomicRegions_FromGFF",
                (comment_char, fix_negative_coordinates),
            ),
            ppg.FunctionInvariant(
                name + "_filter_func_GenomicRegions_FromGFF", filter_function
            ),
            ppg.FunctionInvariant(
                name + "_chromosome_manlger_GenomicRegions_FromGFF", chromosome_mangler
            ),
        ]
    else:
        deps = []
    return alternative_class(
        name, load, deps, genome, on_overlap, summit_annotator=summit_annotator, vid=vid
    )


def GenomicRegions_FromWig(
    name,
    filename,
    genome,
    enlarge_5prime=0,
    enlarge_3prime=0,
    on_overlap="raise",
    comment_char=None,
    summit_annotator=None,
    vid=None,
):
    """Create GenomicRegions from a Wiggle file.

    @enlarge_5prime and @enlarge_3prime increase the size of the fragments described in the wig in
    the respective direction (for example if a chip-chip array did not cover every base).
    @comment_char defines which lines to ignore in the wiggle (see {mbf_fileformats.wiggle_to_intervals})

    The resulting GenomicRegions has a column 'Score' that contains the wiggle score"""
    from mbf_fileformats.wiggle import wiggle_to_intervals

    def load():
        df = wiggle_to_intervals(filename, comment_char=comment_char)
        df["chr"] = [to_string(x) for x in df["chr"]]
        df["start"] -= enlarge_5prime
        df["stop"] += enlarge_3prime
        return df

    if ppg.inside_ppg():
        deps = [ppg.FileTimeInvariant(filename)]
    else:
        deps = []

    return GenomicRegions(
        name, load, deps, genome, on_overlap, summit_annotator=summit_annotator, vid=vid
    )


def GenomicRegions_FromBed(
    name,
    filename,
    genome,
    chromosome_mangler=lambda x: x,
    on_overlap="raise",
    filter_invalid_chromosomes=False,
    summit_annotator=None,
    sheet_name=None,
    vid=None,
):
    """Create GenomicRegions from a Bed file.

    The resulting GenomicRegions has a column 'Score' that contains the wiggle score"""
    from mbf_fileformats.bed import read_bed

    def load():
        valid_chromosomes = set(genome.get_chromosome_lengths())
        data = {}
        entries = read_bed(filename)
        data["chr"] = np.array(
            [chromosome_mangler(to_string(e.refseq)) for e in entries], dtype=np.object
        )
        data["start"] = np.array([e.position for e in entries], dtype=np.int32)
        data["stop"] = np.array(
            [e.position + e.length for e in entries], dtype=np.int32
        )
        data["score"] = np.array([e.score for e in entries], dtype=np.float)
        data["strand"] = np.array([e.strand for e in entries], dtype=np.int8)
        data["name"] = np.array([to_string(e.name) for e in entries], dtype=np.object)
        data = pd.DataFrame(data)
        if filter_invalid_chromosomes:  # pragma: no cover
            keep = [x in valid_chromosomes for x in data["chr"]]
            data = data[keep]
        res = data
        if len(res) == 0:
            raise ValueError("Emtpty Bed file - %s" % filename)
        if (np.isnan(res["score"])).all():
            res = res.drop(["score"], axis=1)
        if (len(res["name"]) > 1) and (len(res["name"].unique()) == 1):
            res = res.drop(["name"], axis=1)
        return res

    if ppg.inside_ppg():
        deps = [
            ppg.FileTimeInvariant(filename),
            ppg.FunctionInvariant(name + "_chrmangler", chromosome_mangler),
        ]
    else:
        deps = []

    return GenomicRegions(
        name,
        load,
        deps,
        genome,
        on_overlap=on_overlap,
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_FromBigBed(
    name,
    filename,
    genome,
    chromosome_mangler=lambda x: x,
    on_overlap="raise",
    summit_annotator=None,
    sheet_name=None,
    vid=None,
):
    """Create GenomicRegions from a BigBed file.
    @chromosome_mangler translates genome chromosomes into the bigbed's chromosomes!

    """
    from mbf_fileformats.bed import read_bigbed

    def load():
        res = read_bigbed(filename, genome.get_chromosome_lengths(), chromosome_mangler)
        if (res["strand"] == 1).all():
            res = res.drop("strand", axis=1)
        if len(res) == 0:  # pragma: no cover
            raise ValueError(
                "Emtpty BigBed file (or wrong chromosome names)- %s" % filename
            )
        return res

    if ppg.inside_ppg():
        deps = [
            ppg.FileTimeInvariant(filename),
            ppg.FunctionInvariant(name + "_chrmangler", chromosome_mangler),
        ]
    else:
        deps = []

    return GenomicRegions(
        name,
        load,
        deps,
        genome,
        on_overlap=on_overlap,
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_BinnedGenome(
    genome, bin_size, limit_to_chromosomes=None, new_name=None, vid=None
):
    """Create GenomicRegions than partition a given chromosome in bins
    of @bin_size bp
    """
    if limit_to_chromosomes and not isinstance(limit_to_chromosomes, list):
        raise ValueError("Limit_to_chromosomes must be a list")

    def load():
        data = {"chr": [], "start": [], "stop": []}
        if limit_to_chromosomes:
            chr_plus_length = [
                (chr, chr_len)
                for (chr, chr_len) in genome.get_chromosome_lengths().items()
                if chr in limit_to_chromosomes
            ]
        else:
            chr_plus_length = genome.get_chromosome_lengths().items()
        if not chr_plus_length:  # pragma: no cover
            raise ValueError("No chromosomes generated")
        for chr, chr_len in chr_plus_length:
            no_of_bins = int(math.ceil(float(chr_len) / bin_size))
            data["chr"].extend([chr] * no_of_bins)
            starts = np.array(range(0, chr_len, bin_size))
            stops = starts + bin_size  # not -1, we are right exclusive!
            data["start"].extend(starts)
            data["stop"].extend(stops)
        res = pd.DataFrame(data)
        return res

    if new_name is None:
        name = genome.name + " binned (%i bp)" % bin_size
    else:
        name = new_name
    return GenomicRegions(name, load, [], genome, vid=vid)


def GenomicRegions_Union(
    name, list_of_grs, summit_annotator=None, sheet_name="Overlaps"
):
    """Combine serveral GRs into one.

    Do not set on_overlap


    """
    verify_same_genome(list_of_grs)

    def load():
        dfs = [x.df[["chr", "start", "stop"]] for x in list_of_grs]
        return pd.concat(dfs, axis=0)

    if ppg.inside_ppg():
        deps = [x.load() for x in list_of_grs]
        deps.append(
            ppg.ParameterInvariant(
                name + "_input_grs", list(sorted([x.name for x in list_of_grs]))
            )
        )
    else:
        deps = []
    vid = ("union", [x.vid for x in list_of_grs])
    return GenomicRegions(
        name,
        load,
        deps,
        list_of_grs[0].genome,
        on_overlap="merge",
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_Overlapping(
    new_name, gr_a, gr_b, summit_annotator=None, sheet_name="Overlaps"
):
    return GenomicRegions_Common(new_name, [gr_a, gr_b], summit_annotator, sheet_name)


def GenomicRegions_Common(
    name, list_of_grs, summit_annotator=None, sheet_name="Overlaps"
):
    """Combine serveral GRs into one. Keep only those (union) regions occuring in all."""

    def load():
        union = merge_df_intervals(
            pd.concat([x.df[["chr", "start", "stop"]] for x in list_of_grs])
        ).reset_index(drop=True)
        keep = np.ones((len(union),), dtype=np.bool)
        for gr in list_of_grs:
            for ii, row in union.iterrows():
                if keep[
                    ii
                ]:  # no point in checking if we already falsified - short circuit...
                    if not gr.has_overlapping(row["chr"], row["start"], row["stop"]):
                        keep[ii] = False
        return union[keep]

    verify_same_genome(list_of_grs)
    if ppg.inside_ppg():
        deps = [x.load() for x in list_of_grs]
        deps.append(
            ppg.ParameterInvariant(
                name + "_input_grs", sorted([x.name for x in list_of_grs])
            )
        )
    else:
        for x in list_of_grs:
            x.load()
        deps = []
    vid = ("common", [x.vid for x in list_of_grs])
    return GenomicRegions(
        name,
        load,
        deps,
        list_of_grs[0].genome,
        on_overlap="raise",
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_Invert(new_name, gr, summit_annotator=None, sheet_name="Inverted"):
    """Invert a GenomicRegions. What was covered becomes uncovered, what was uncovered becomes covered.
        [(10, 100), (400, 450)], in a chromosome of size 1000
        becomes
        [(0, 10), (450, 1000)]

        Note that all interval-set based operations (L{union}, L{intersection}, L{difference})
        drop all columns but chr, start, stop (annotators are merged and readded from all sets involved)
        """

    def do_load():
        from mbf_nested_intervals import IntervalSet
        import itertools

        df = gr.df
        joined = []
        for ii, (chr, start, stop) in enumerate(
            zip(df["chr"], df["start"], df["stop"])
        ):
            joined.append(((chr), start, stop, ii))
        joined.sort(key=lambda tup: tup[0])

        out = []
        chr_lengths = gr.genome.get_chromosome_lengths()
        seen = set()
        for chr, sub_group in itertools.groupby(joined, lambda tup: tup[0]):
            args = [x[1:] for x in sub_group]
            iv = IntervalSet.from_tuples_with_id(args)
            new_order = iv.invert(0, chr_lengths[chr]).to_numpy()
            out.append(
                pd.DataFrame({"start": new_order[0], "stop": new_order[1], "chr": chr})
            )
            seen.add(chr)
        for chr in chr_lengths.keys() - seen:
            out.append(
                pd.DataFrame({"start": [0], "stop": [chr_lengths[chr]], "chr": chr})
            )

        return pd.concat(out).reset_index(drop=True)

    if gr.load_strategy.build_deps:
        deps = [
            gr.load(),
            ppg.ParameterInvariant(
                "GenomicRegions_%s_parents" % new_name, (gr.name)
            ),  # so if you swap out the gr, it's detected...
        ]
    else:
        deps = []

    result = GenomicRegions(
        new_name,
        do_load,
        deps,
        gr.genome,
        on_overlap="raise",
        summit_annotator=summit_annotator,
        vid=(["invert"] + gr.vid) if gr.vid is not None else None,
        sheet_name=sheet_name,
    )
    return result


def GenomicRegions_Difference(
    new_name, gr_a, gr_b, summit_annotator=None, sheet_name="difference"
):
    """Create a difference of these intervals wth other_gr's intervalls
        [(10, 100), (400, 450)],
        [(80, 120), (600, 700)]
        becomes
        [(10, 80), (400, 450)]
        (intervals may be split up!)

        Note that all interval-set based operations (L{union}, L{intersection}, L{difference})
        drop all columns but chr, start, stop (annotators are merged and readded from all sets involved)
        """
    verify_same_genome([gr_a, gr_b])

    def do_load():
        new_rows = []
        for idx, row in gr_a.df[["chr", "start", "stop"]].iterrows():
            overlaps = gr_b.get_overlapping(row["chr"], row["start"], row["stop"])
            if not len(overlaps):  # the easy case...
                new_rows.append(row)
            else:
                overlaps = merge_df_intervals(
                    overlaps
                )  # they are now also sorted, so all we need to do is walk them, keep the regions between them (and within the original interval)
                start = row["start"]
                if overlaps.at[0, "start"] <= start and overlaps.at[0, "stop"] > start:
                    start_i = 1
                    start = overlaps.at[0, "stop"]
                else:
                    start_i = 0
                for ii in range(start_i, len(overlaps)):
                    stop = min(overlaps.at[ii, "start"], row["stop"])
                    new_rows.append({"chr": row["chr"], "start": start, "stop": stop})
                    start = overlaps.at[ii, "stop"]
                if start < row["stop"]:
                    new_rows.append(
                        {"chr": row["chr"], "start": start, "stop": row["stop"]}
                    )

                # todo: work the cutting up magic!
                pass
        if new_rows:
            return pd.DataFrame(new_rows)
        else:
            return pd.DataFrame({"chr": [], "start": [], "stop": []})

    if gr_a.load_strategy.build_deps:
        deps = [
            gr_b.load(),
            gr_a.load(),
            ppg.ParameterInvariant(
                "GenomicRegions_%s_parents" % new_name, (gr_a.name, gr_b.name)
            ),  # so if you swap out the gr, it's detected...
        ]
    else:
        gr_b.load()
        deps = []

    result = GenomicRegions(
        new_name,
        do_load,
        deps,
        gr_a.genome,
        on_overlap="merge",
        summit_annotator=summit_annotator,
        vid=["difference"] + list(gr_a.vid) + list(gr_b.vid),
        sheet_name=sheet_name,
    )
    return result


def GenomicRegions_Intersection(
    new_name, gr_a, gr_b, summit_annotator=None, sheet_name="intersection"
):
    """Create an intersection of all intervals...
        [(10, 100), (400, 450)],
        [(80, 120), (600, 700)]
        becomes
        [(80, 100),]

        Note that all interval-set based operations (L{union}, L{intersection}, L{difference})
        drop all columns but chr, start, stop (annotators are merged and readded from all sets involved)
        """
    verify_same_genome([gr_a, gr_b])

    def do_load():
        new_rows = []
        for chr, start, stop in gr_a._iter_intersections(gr_b):
            new_rows.append({"chr": chr, "start": start, "stop": stop})
        if new_rows:
            return pd.DataFrame(new_rows)
        else:
            return pd.DataFrame({"chr": [], "start": [], "stop": []})

    if gr_a.load_strategy.build_deps:
        deps = [
            gr_b.load(),
            gr_a.load(),
            ppg.ParameterInvariant(
                "GenomicRegions_%s_parents" % new_name, (gr_a.name, gr_b.name)
            ),  # so if you swap out the gr, it's detected...
        ]
    else:
        deps = []
        gr_b.load()

    result = GenomicRegions(
        new_name,
        do_load,
        deps,
        gr_a.genome,
        on_overlap="merge",
        summit_annotator=summit_annotator,
        vid=["intersection"] + gr_a.vid + gr_b.vid,
        sheet_name=sheet_name,
    )
    return result


def GenomicRegions_FromPartec(
    name, filename, genome, on_overlap="raise", summit_annotator=None, vid=None
):
    """create GenomicRegions from Partec's output"""
    return GenomicRegions_FromTable(
        name,
        filename,
        genome,
        on_overlap=on_overlap,
        summit_annotator=summit_annotator,
        vid=vid,
        chr_column="Chromosome",
        start_column="Start",
        stop_column="Stop",
        drop_further_columns=False,
        reader=lambda x: pd.read_csv(x, sep="\t"),
    )


def GenomicRegions_FromTable(
    name,
    filename,
    genome,
    on_overlap="raise",
    summit_annotator=None,
    filter_func=None,
    vid=None,
    sheet_name="FromTable",
    drop_further_columns=True,
    chr_column="chr",
    start_column="start",
    stop_column="stop",
    one_based=False,
    reader=read_pandas,
):
    """Read a table file (csv/tsv/xls) with the chr/start/stop columns (renamed?), optionally
    drop all further columns"""

    def load():

        df = reader(filename)
        df["chr"] = df[chr_column].astype(str)
        df["start"] = df[start_column].astype(int)
        if one_based:  # pragma: no cover
            df["start"] -= 1
        df["stop"] = df[stop_column].astype(int)
        if drop_further_columns:  # pragma: no cover
            df = df[["chr", "start", "stop"]]
        if filter_func:  # pragma: no cover
            df = filter_func(df)
        return df

    if ppg.inside_ppg():
        deps = [
            ppg.FileTimeInvariant(filename),
            ppg.FunctionInvariant(name + "_filter_func", filter_func),
        ]
    else:
        deps = []
    return GenomicRegions(
        name,
        load,
        deps,
        genome,
        on_overlap,
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_CommonInAtLeastX(
    name, list_of_grs, X, summit_annotator=None, sheet_name="Overlaps"
):
    """Combine serveral GRs into one. Keep only those (union) regions occuring in at least x."""

    def load():
        union = merge_df_intervals(
            pd.concat([x.df[["chr", "start", "stop"]] for x in list_of_grs])
        ).reset_index()
        keep = np.zeros((len(union),), dtype=np.bool)
        for ii, row in union.iterrows():
            count = 0
            for gr in list_of_grs:
                if gr.has_overlapping(row["chr"], row["start"], row["stop"]):
                    count += 1
            keep[ii] = count >= X
        if not keep.any():  # pragma: no cover
            raise ValueError("Filtered all of them")
        return union.iloc[keep]

    if len(set([x.genome for x in list_of_grs])) > 1:  # pragma: no cover
        raise ValueError("Can only merge GenomicRegions that have the same genome")
    if ppg.inside_ppg():
        deps = [x.load() for x in list_of_grs]
        deps.append(
            ppg.ParameterInvariant(
                name + "_input_grs", sorted([x.name for x in list_of_grs])
            )
        )
    else:
        deps = []
        [x.load() for x in list_of_grs]
    vid = ("common at least %i" % X, [x.vid for x in list_of_grs])
    return GenomicRegions(
        name,
        load,
        deps,
        list_of_grs[0].genome,
        on_overlap="raise",
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
        vid=vid,
    )


def GenomicRegions_Windows(
    genome,
    name,
    window_size,
    window_spacing,
    subset_of_chromosomes=None,
    sheet_name="Windowed",
):
    """Create a GenomicRegions that has a window of size @window_size (0 for next to each other), windows are spaced @window_spacing
    across all or a @subset_of_chromosomes"""
    if isinstance(genome, str):  # pragma: no cover
        raise TypeError("Wrong arguments, genome must be first")

    def load():
        chrs_to_include = list(
            subset_of_chromosomes
            if subset_of_chromosomes
            else genome.get_chromosome_lengths().keys()
        )
        chrs = []
        starts = []
        stops = []
        chr_lengths = genome.get_chromosome_lengths()
        for c in chrs_to_include:
            ll = chr_lengths[c]
            for ii in range(0, ll, window_size + window_spacing):
                chrs.append(c)
                starts.append(ii)
                stops.append(ii + window_size)
        return pd.DataFrame({"chr": chrs, "start": starts, "stop": stops})

    return GenomicRegions(
        name, load, [], genome, on_overlap="raise", sheet_name=sheet_name
    )


def GenomicRegions_FilterRemoveOverlapping(
    new_name, gr_a, other_grs, summit_annotator=None, sheet_name="Overlaps"
):
    """Filter all from this GenomicRegions that have an overlapping region in other_gr
    Note that filtering does not change the coordinates, it only filters,
    non annotator additional rows are kept, annotators are recalculated.
    """
    if isinstance(other_grs, GenomicRegions):
        other_grs = [other_grs]
    verify_same_genome([gr_a] + other_grs)

    def filter_func(df):
        keep = np.zeros((len(df)), dtype=np.bool)
        for ii, tup in enumerate(df[["chr", "start", "stop"]].itertuples()):
            for other_gr in other_grs:
                keep[ii] = keep[ii] | other_gr.has_overlapping(
                    tup.chr, tup.start, tup.stop
                )
        return ~keep

    if not summit_annotator:  # pragma: no cover
        summit_annotator = gr_a.summit_annotator

    if gr_a.load_strategy.build_deps:
        deps = [
            [x.load() for x in other_grs],
            ppg.ParameterInvariant(
                "GenomicRegions_%s_parents" % new_name,
                (gr_a.name, [x.name for x in other_grs]),
            ),  # so if you swap out the gr, it's detected...
        ]
    else:
        for other_gr in other_grs:
            other_gr.load()
        deps = []

    return gr_a.filter(
        new_name,
        df_filter_function=filter_func,
        dependencies=deps,
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
    )


def GenomicRegions_FilterToOverlapping(
    new_name, gr_a, other_grs, summit_annotator=None, sheet_name="Overlaps"
):
    """Filter to just those that overlap one in *all* other_grs.
    Note that filtering does not change the coordinates, it only filters,
    non annotator additional rows are kept, annotators are recalculated.
    """
    if isinstance(other_grs, GenomicRegions):
        other_grs = [other_grs]
    verify_same_genome([gr_a] + other_grs)

    def filter_func(df):
        keep = np.ones((len(df)), dtype=np.bool)
        for ii, row in enumerate(df[["chr", "start", "stop"]].itertuples()):
            for gr in other_grs:
                keep[ii] &= gr.has_overlapping(row.chr, row.start, row.stop)
        return keep

    if gr_a.load_strategy.build_deps:
        deps = [other_gr.load() for other_gr in other_grs] + [
            ppg.ParameterInvariant(
                "GenomicRegions_%s_parents" % new_name,
                (gr_a.name, [other_gr.name for other_gr in other_grs]),
            )  # so if you swap out the gr, it's detected...
        ]
    else:
        for other_gr in other_grs:
            other_gr.load()
        deps = []

    return gr_a.filter(
        new_name,
        df_filter_function=filter_func,
        dependencies=deps,
        summit_annotator=summit_annotator,
        sheet_name=sheet_name,
    )



from .genes import Genes
from .genes_from import (
    FromDifference,
    FromIntersection,
    FromAny,
    FromAll,
    FromNone,
    FromFile,
    FromFileOfTranscripts,
    FromBiotypes,
    FromNames,
)
from . import anno_tag_counts
from . import annotators


__all__ = [
    "Genes",
    "FromDifference",
    "FromIntersection",
    "FromAny",
    "FromAll",
    "FromNone",
    "FromFile",
    "FromFileOfTranscripts",
    "FromBiotypes",
    "FromNames",
    "anno_tag_counts",
    "annotators",
]



from mbf_genomes import HardCodedGenome


def MockGenome(  # noqa: C901
    df_genes, df_transcripts=None, chr_lengths=None, df_genes_meta=None
):  # noqa: C901
    if chr_lengths is None:
        chr_lengths = {
            "1": 100_000,
            "2": 200_000,
            "3": 300_000,
            "4": 400_000,
            "5": 500_000,
        }

    df_genes = df_genes.rename(columns={"stable_id": "gene_stable_id"})
    if not "start" in df_genes.columns:
        starts = []
        stops = []
        if not "strand" in df_genes:
            tes_larger = df_genes["tes"] > df_genes["tss"]
            strand = tes_larger.replace({True: 1, False: -1})
            df_genes = df_genes.assign(strand=strand)
        for idx, row in df_genes.iterrows():
            starts.append(min(row["tss"], row["tes"]))
            stops.append(max(row["tss"], row["tes"]))
        df_genes = df_genes.assign(start=starts, stop=stops)
    if not "biotype" in df_genes.columns:
        df_genes = df_genes.assign(biotype="protein_coding")
    if not "name" in df_genes.columns:
        df_genes = df_genes.assign(name=df_genes.gene_stable_id)
    df_genes = df_genes.sort_values(["chr", "start"])
    df_genes = df_genes.set_index("gene_stable_id")
    if not df_genes.index.is_unique:  # pragma: no cover
        raise ValueError("gene_stable_ids not unique")
    if df_transcripts is not None and len(df_transcripts):
        if not "transcript_stable_id" in df_transcripts.columns:
            print(df_transcripts.columns)
            df_transcripts = df_transcripts.assign(
                transcript_stable_id=df_transcripts["name"]
            )
        if not "biotype" in df_transcripts.columns:
            df_transcripts = df_transcripts.assign(biotype="protein_coding")
        if not "name" in df_transcripts.columns:
            df_transcripts = df_transcripts.assign(
                name=df_transcripts.transcript_stable_id
            )
        if "exons" in df_transcripts.columns:
            if len(df_transcripts["exons"].iloc[0]) == 3:  # pragma: no cover
                df_transcripts = df_transcripts.assign(
                    exons=[(x[0], x[1]) for x in df_transcripts["exons"]]
                )
            df_transcripts = df_transcripts.assign(
                exon_stable_ids=[
                    "exon_%s_%i" % (idx, ii)
                    for (ii, idx) in enumerate(df_transcripts["exons"])
                ]
            )
        stops = []
        if not "strand" in df_transcripts:  # pragma: no cover
            df_transcripts = df_transcripts.assign(strand=1)
        if not "tss" in df_transcripts:  # pragma: no branch
            tss = []
            tes = []
            for tr, row in df_transcripts.iterrows():
                if row["strand"] == 1:
                    tss.append(min((x[0] for x in row["exons"])))
                    tes.append(max((x[1] for x in row["exons"])))
                else:
                    tss.append(max((x[1] for x in row["exons"])))
                    tes.append(min((x[0] for x in row["exons"])))
            df_transcripts = df_transcripts.assign(tss=tss, tes=tes)
        if not "start" in df_transcripts:
            starts = []
            stops = []
            for idx, row in df_transcripts.iterrows():
                starts.append(min(row["tss"], row["tes"]))
                stops.append(max(row["tss"], row["tes"]))
            df_transcripts = df_transcripts.assign(start=starts, stop=stops)
        df_transcripts = df_transcripts.set_index("transcript_stable_id")
        if not df_transcripts.index.is_unique:  # pragma: no cover
            raise ValueError("transcript_stable_ids not unique")
    result = HardCodedGenome("dummy", chr_lengths, df_genes, df_transcripts, None)
    result.sanity_check_genes(df_genes)
    if df_transcripts is not None and len(df_transcripts):
        result.sanity_check_transcripts(df_transcripts)
    if df_genes_meta is not None:
        result.df_genes_meta = df_genes_meta
    return result



import numpy as np
import pypipegraph as ppg
from .genes import Genes
from ..util import read_pandas, freeze


def FromDifference(name, a, b, sheet_name="Differences"):

    """a minus b"""

    def do_load(df):
        remove_ids = set(b.df["gene_stable_id"])
        keep = ~np.array(
            [stable_id in remove_ids for stable_id in a.df["gene_stable_id"]],
            dtype=np.bool,
        )
        return keep

    if a.load_strategy.build_deps:
        deps = [
            a.load(),
            b.load(),
            ppg.ParameterInvariant(
                "Genes_%s_parents" % name, (a.name, b.name)
            ),  # so if you swap out the gr, it's detected...
        ]
    else:
        deps = []

    res = a.filter(
        name,
        do_load,
        dependencies=deps,
        sheet_name=sheet_name,
        vid=["Difference", a.vid, b.vid],
    )
    res.parent = a
    return res


def FromIntersection(name, gene_sets, sheet_name="Intersections"):
    if not isinstance(sheet_name, str):
        raise ValueError(
            "sheet name must be string - note that FromIntersection"
            "takes a list of genes as 2nd parameter"
        )
    parents = []
    vid = _vid_from_genes_sets(gene_sets, "Intersection")
    for g in gene_sets:
        parent = g.parent
        while parent.parent:
            parent = parent.parent
        parents.append(parent)
    parent_names = set([x.name for x in parents])
    if len(parent_names) != 1:  # pragma: no cover
        raise ValueError(
            "Trying to combine gene lists from different parents - currently not supported"
        )

    def in_set(df):
        accepted_ids = set.intersection(
            *[set(g.df["gene_stable_id"]) for g in gene_sets]
        )
        return np.array(
            [x in accepted_ids for x in df["gene_stable_id"]], dtype=np.bool
        )

    return parents[0].filter(
        name,
        in_set,
        dependencies=[g.load() for g in gene_sets],
        sheet_name=sheet_name,
        vid=vid,
    )


def _vid_from_genes_sets(genes_sets, prefix="filtered", sheet_name=None):
    if sheet_name is not None:
        vid = [prefix, sheet_name]
    else:
        vid = [prefix]
    for g in genes_sets:
        if g.vid is not None:
            if isinstance(g.vid, str):
                vid.append(g.vid)
            else:
                vid = vid + g.vid
    return vid


def _from_filtered_genes(name, do_filter, genes_sets, sheet_name, vid):
    if genes_sets[0].load_strategy.build_deps:
        deps = ([o.load() for o in genes_sets],)
    else:
        deps = []
    return Genes(genes_sets[0].genome).filter(
        name, do_filter, dependencies=deps, sheet_name=sheet_name, vid=vid
    )


def FromAny(name, genes_sets, sheet_name=None):
    def do_filter(df):
        seen = set()
        for o in genes_sets:
            seen.update(o.df["gene_stable_id"])
        return np.array([stable_id in seen for stable_id in df["gene_stable_id"]])

    return _from_filtered_genes(
        name,
        do_filter,
        genes_sets,
        sheet_name,
        vid=_vid_from_genes_sets(genes_sets, "filtered", sheet_name),
    )


def FromAll(name, genes_sets, sheet_name=None):
    # def filter_to_those_occuring_in_all_filtered_sets(
    def do_filter(df):
        ok = set.intersection(
            *[set(o.df["gene_stable_id"].unique()) for o in genes_sets]
        )
        return np.array([stable_id in ok for stable_id in df["gene_stable_id"]])

    return _from_filtered_genes(
        name,
        do_filter,
        genes_sets,
        sheet_name,
        vid=_vid_from_genes_sets(genes_sets, "filtered", sheet_name),
    )


def FromNone(name, genes_sets, sheet_name=None):
    # def filter_to_those_not_occuring_in_any_filtered_sets(
    def do_filter(df):
        seen = set()
        for o in genes_sets:
            seen.update(o.df["gene_stable_id"])
        return np.array(
            ~np.array(
                [stable_id in seen for stable_id in df["gene_stable_id"]], dtype=np.bool
            )
        )

    return _from_filtered_genes(
        name,
        do_filter,
        genes_sets,
        sheet_name,
        vid=_vid_from_genes_sets(genes_sets, "filtered", sheet_name),
    )


def FromFile(
    name, genome, table_filename, column_name="gene_stable_id", sheet_name=None
):
    """Filter Genes(genome) to those occuring in the table_filename"""

    table_filename = str(table_filename)

    def filter(genes_df):

        df = read_pandas(table_filename)
        seen = df[column_name].values
        return np.array(
            [str(x) in seen for x in genes_df["gene_stable_id"]], dtype=np.bool
        )

    g = Genes(genome)
    if g.load_strategy.build_deps:
        deps = [ppg.FileChecksumInvariant(table_filename)]
    else:
        deps = []

    return g.filter(name, filter, dependencies=deps, sheet_name=sheet_name)


def FromFileOfTranscripts(
    name, genome, table_filename, column_name="transcript_stable_id"
):
    """Filter Genes(genome) to those whose transcripts occur in the table_filename"""

    def filter(genes_df):
        df = read_pandas(table_filename)
        transcripts = df[column_name]
        seen = set()
        for transcript_stable_id in transcripts:
            seen.add(genome.transcripts[transcript_stable_id].gene_stable_id)

        return np.array([x in seen for x in genes_df["gene_stable_id"]], dtype=np.bool)

    g = Genes(genome)
    if g.load_strategy.build_deps:
        deps = [ppg.FileChecksumInvariant(table_filename)]
    else:
        deps = []
    return g.filter(name, filter, dependencies=deps)


def FromBiotypes(genome, allowed_biotypes):
    def filter(genes_df):
        ok = np.zeros(genes_df.shape[0], dtype=np.bool)
        for x in allowed_biotypes:
            ok = ok | (genes_df["biotype"] == x)
        return ok

    return Genes(genome).filter(
        "%s_with_%s" % (genome.name, ",".join(allowed_biotypes)), filter
    )


def FromNames(
    name,
    genome,
    list_or_callback,
    sheet_name=None,
    vid=None,
    manual_lookup=None,
    ignore_unmatched=False,
):
    """Filter Genes(genome) to those occuring in the list (or the list
    returned from the callback.

    If manual_lookup is set, genes will be first replaced with those before lookup
    in the genome. Entries resolving to none will be excluded.
    Manual lookup may contain unused entries

    """

    def filter(genes_df, manual_lookup=manual_lookup):
        if hasattr(list_or_callback, "__call__"):
            ll = list_or_callback()
        else:
            ll = list_or_callback
        if manual_lookup is not None:
            manual_lookup = {
                k.upper(): v.upper() if v is not None else None
                for (k, v) in manual_lookup.items()
            }
            ll = [manual_lookup.get(x.upper(), x.upper()) for x in ll]
            ll = [x for x in ll if x is not None]

        stable_ids = set(genes_df["gene_stable_id"])
        ids_seen = [x for x in ll if x in stable_ids]
        names = [x for x in ll if x not in stable_ids]
        seen = set([x.upper() for x in names])
        unused = seen.difference(genes_df["name"].str.upper())
        if unused and not ignore_unmatched:
            raise ValueError(
                "the following gene names were not found in the genome:\n"
                + "\n".join(sorted(unused))
            )
        res = np.zeros(len(genes_df), bool)
        res |= genes_df["gene_stable_id"].isin(ids_seen)
        res |= genes_df["name"].str.upper().isin(seen)
        return res

    g = Genes(genome)
    if g.load_strategy.build_deps:
        if hasattr(list_or_callback, "__call__"):
            deps = [ppg.FunctionInvariant(name + "_list_or_callback", list_or_callback)]
        else:
            deps = [
                ppg.ParameterInvariant(name + "_list_or_callback", list_or_callback)
            ]
        deps.append(
            ppg.ParameterInvariant(name + "_manual_lookup", freeze(manual_lookup))
        )
    else:
        deps = []

    return g.filter(name, filter, dependencies=deps, sheet_name=sheet_name, vid=vid)



import pypipegraph as ppg
import numpy as np
import pandas as pd
from pathlib import Path
import itertools
from collections.abc import Iterator


from mbf_genomics.delayeddataframe import DelayedDataFrame
from mbf_genomes import GenomeBase
from mbf_externals.util import lazy_property
from mbf_nested_intervals import (
    merge_df_intervals,
    merge_df_intervals_with_callback,
    IntervalSet,
    _df_to_tup_no_strand,
)


def merge_identical_but_raise_on_further_overlap(df):
    def iv_func(iv):
        iv = iv.remove_duplicates()
        if iv.any_overlapping():
            raise ValueError(
                "Overlapping after merging identical present. Use  a different on_overlap or fix our data"
            )
        return iv

    res = merge_df_intervals(df, iv_func)
    return res


region_registry = {}


class GenomicRegions(DelayedDataFrame):

    """A genomic regions object encapsulates intervals across the chromosomes
    of a genome. They model everything from restriction sites over binding sites
    to gene locations.

    Internally a genomic region consists of a dataframe {chr, start, stop} (.df),
    and GR-annotators can be used to add further columns.
    Please note that the intervals are 'standard python' ie. left inclusive, right exclusive.
    An interval 10:20, is 10 bp long, and covers 10, 11, ..., 18, 19.
    That also means a 'point' interval is actually x, x+1, not x,x (that's an empty interval)

    There is a variety of set operations you can perform on them that are done efficiently
    via interval trees.

    GRs do not require disjoint regions. More specifically, they can be configured
    to do a variety of things if they're passed overlapping regions - ignore them,
    raise an exception, merge them...

    You can query them for all of their internal regions overlapping an interval

    Annotators are inherited by descendands, their values are recalculated for
    the new intervals.

    """

    # basic loading
    def __init__(
        self,
        name,
        loading_function,
        dependencies,
        genome,
        on_overlap="raise",
        result_dir=None,
        summit_annotator=None,
        sheet_name=None,
        vid=None,
    ):
        """Create a lazy loaded GenomicRegion, available once it's load() job has been completed.
        @dependencies allow you to inject dependencies for the load() job.
        @on_overlap may be one of (raise, ignore, merge, drop, see below), which decides how overlapping regions are handled.

        @loading_function is a parameterless function that returns dataframe of {chr, start, stop}.
        The dataframe does not need to sort the data in any way. chr must be a string column, start and stop integers
        (this is checked).

        The internal df is always sorted by chr, start...


        @on_overlap, allowed values:
            -raise: - raise a ValueError if any overlapping regions are detected (default)
            -merge: - combine overlapping regions into larger ones
            -merge_identical: combine identical regions, but raise if there are non identical overlapping ones
            -(merge, merge_function) - combine regions - call merge_function(sub_df) which must return a row (dict) to keep and is only called when there are multiple rows to pick! Its row['start'], row['stop'] are ignored though
            -drop: - drop all regions that are overlapping one another
            -ignore: - ignore overlapping regions (this implies a nested list search for overlap queries
            and will lead to certain functions that assume non-overlapping regions (such as covered_bases)
            to raise ValueErrors)

        """
        if (
            isinstance(dependencies, str)
            or not hasattr(dependencies, "__iter__")
            or (
                hasattr(dependencies, "__iter__") and isinstance(dependencies, Iterator)
            )
        ):
            raise ValueError(
                "dependencies must be iterable (use [] for no external dependencies"
            )
        if not isinstance(dependencies, list):
            dependencies = list(dependencies)

        allowed_overlap_modes = ("raise", "merge", "ignore", "drop", "merge_identical")
        if not on_overlap in allowed_overlap_modes and not (
            isinstance(on_overlap, tuple)
            and on_overlap[0] == "merge"
            and hasattr(on_overlap[1], "__call__")
        ):
            raise ValueError(
                "Invalid on_overlap mode %s. Allowed: %s, or a tuple of ('merge', function)"
                % (on_overlap, allowed_overlap_modes)
            )

        self.name = name
        self.gr_loading_function = loading_function
        self.genome = genome
        self.on_overlap = on_overlap
        self._default_mangler = True
        if self.on_overlap == "ignore":
            self.need_to_handle_overlapping_regions = True
        elif (
            self.on_overlap == "raise"
            or self.on_overlap == "merge"
            or self.on_overlap == "merge_identical"
            or self.on_overlap == "drop"
            or (
                isinstance(self.on_overlap, tuple)
                and self.on_overlap[0] == "merge"
                and hasattr(self.on_overlap[1], "__call__")
            )
        ):
            self.need_to_handle_overlapping_regions = False
        else:  # pragma: no cover - defensive
            raise ValueError(
                "Don't know how to decide on has_overlapping from %s"
                % (self.on_overlap,)
            )

        if result_dir:
            result_dir = Path(result_dir)
        else:
            if sheet_name:
                result_dir = Path("results") / "GenomicRegions" / sheet_name / name
            else:
                result_dir = Path("results") / "GenomicRegions" / name
        self.sheet_name = sheet_name

        super().__init__(name, self._load, dependencies, result_dir)
        if self.load_strategy.build_deps:
            dependencies = []
            if isinstance(on_overlap, tuple):
                dependencies.append(
                    ppg.ParameterInvariant("grload_params" + name, ("function",))
                )
                dependencies.append(
                    ppg.FunctionInvariant("grload_overlap_func" + name, on_overlap[1])
                )
            else:
                dependencies.append(
                    ppg.ParameterInvariant("grload_params" + name, (on_overlap,))
                )
            dependencies.extend(genome.download_genome())
            dependencies.append(
                ppg.FunctionInvariant("grload_func" + name, self.gr_loading_function)
            )
            self.load_strategy.load().depends_on(dependencies)

        if not hasattr(self, "column_properties"):
            self.column_properties = {
                "chr": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "On which chromosome (or contig) the region is loacted",
                },
                "start": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Left most position of this region",
                },
                "stop": {
                    "user_visible": False,
                    "priority": -997,
                    "description": "Right most position of this region",
                },
            }
        # self.genome = genome

        self.random_count = 0
        if summit_annotator:
            self.summit_annotator = summit_annotator
        elif summit_annotator is False:
            self.summit_annotator = None
        else:
            from .annotators import SummitMiddle
            self.summit_annotator = SummitMiddle()
        if self.summit_annotator is not None:
            self.add_annotator(self.summit_annotator)
        self.register()
        if vid is None:
            vid = []
        self.vid = vid

    def get_default_columns(self):
        return ("chr", "start", "stop")

    def _load(self):

        df = self.gr_loading_function()
        if not isinstance(df, pd.DataFrame):
            raise ValueError(
                "GenomicRegion(%s).loading_function must return a pandas.DataFrame, was: %s\n%s"
                % (self.name, type(df), self.gr_loading_function)
            )
        for col in self.get_default_columns():
            if not col in df.columns:
                raise ValueError(
                    "%s not in dataframe returned by GenomicRegion.loading_function"
                    % col
                )
        allowed_chromosomes = set(self.genome.get_chromosome_lengths().keys())
        if len(df):
            # if isinstance(df.iloc[0]["chr"], six.string_types):
            # df["chr"] = [str(x) for x in df["chr"].values]
            if not (isinstance(df.iloc[0]["chr"], str)):
                raise ValueError(
                    "Chromosomes must be a string, was: %s, first instance %s of type %s"
                    % (
                        df["chr"].dtype,
                        repr(df.iloc[0]["chr"]),
                        type(df.iloc[0]["chr"]),
                    )
                )
            chrs = set([str(x) for x in df["chr"].unique()])
            if chrs.difference(allowed_chromosomes):
                raise ValueError(
                    "Invalid chromosomes found: %s, expected one of: %s"
                    % (
                        chrs.difference(allowed_chromosomes),
                        sorted(allowed_chromosomes),
                    )
                )
            if not np.issubdtype(df["start"].dtype, np.integer):
                raise ValueError(
                    "start needs to be an integer, was: %s" % df["start"].dtype
                )
            if not np.issubdtype(df["stop"].dtype, np.integer):
                raise ValueError(
                    "stop needs to be an integer, was: %s" % df["stop"].dtype
                )
            if (df["start"] < 0).any():
                print(df[df["start"] < 0])
                raise ValueError("All starts need to be positive!")
            if (df["start"] > df["stop"]).any():
                error = df[(df["start"] > df["stop"])][["chr", "start", "stop"]]
                error = error[:100]  # not too many...
                error = str(error)
                raise ValueError(
                    "%s.loading_function returned a negative interval:\n%s"
                    % (self, error)
                )
            df = self.handle_overlap(df)
        else:
            df = df.assign(
                start=np.array([], dtype=np.int32),
                stop=np.array([], dtype=np.int32),
                chr=np.array([], dtype=np.object),
            )
        # enforce column order
        cols = ["chr", "start", "stop"]
        for x in df.columns:
            if not x in cols:
                cols.append(x)
        df = df[cols]

        return df

    def handle_overlap(self, df):
        """depending on L{GenomicRegion.on_overlap}, check
        for overlapping regions and handle accordingly"""
        if self.on_overlap == "raise":
            df = df.sort_values(["chr", "start"], ascending=[True, True])
            last_chr = None
            last_stop = 0
            last_row = None
            for idx, row in df.iterrows():
                if row["chr"] != last_chr:
                    last_chr = row["chr"]
                    last_stop = 0
                if row["start"] < last_stop:
                    raise ValueError(
                        "%s: Overlapping intervals: %s %i..%i vs %i..%i"
                        % (
                            self.name,
                            row["chr"],
                            row["start"],
                            row["stop"],
                            last_row["start"],
                            last_row["stop"],
                        )
                    )
                last_stop = row["stop"]
                last_row = row
            return df
        elif self.on_overlap == "merge":
            return merge_df_intervals(df)
        elif self.on_overlap == "merge_identical":
            return merge_identical_but_raise_on_further_overlap(df)
        elif (
            isinstance(self.on_overlap, tuple)
            and self.on_overlap[0] == "merge"
            and hasattr(self.on_overlap[1], "__call__")
        ):
            return merge_df_intervals_with_callback(df, self.on_overlap[1])

        elif self.on_overlap == "ignore":
            df = df.sort_values(["chr", "start"], ascending=[True, True])
            ov = []
            tups = _df_to_tup_no_strand(df)
            for chr, sub_group in itertools.groupby(tups, lambda tup: tup[0]):
                args = [x[1:] for x in sub_group]
                iv = IntervalSet.from_tuples_with_id(args)
                ov.extend(iv.overlap_status())
            df = df.assign(is_overlapping=ov)
            return df

        elif self.on_overlap == "drop":
            return merge_df_intervals(df, lambda iv: iv.merge_drop())
        else:  # pragma: no branch - defensive
            raise NotImplementedError(  # pragma: no cover
                "This branch should not happen - unhandled on_overlap"
            )

    def do_build_intervals(self):
        """"Build the interval trees right now, ignoring all dependencies"""
        if not hasattr(self, "_interval_sets"):
            self._interval_sets = {}
            for chr, tups in itertools.groupby(
                self.df[["chr", "start", "stop"]].itertuples(), key=lambda tup: tup.chr
            ):
                tups = [(tup.start, tup.stop, tup[0]) for tup in tups]
                self._interval_sets[chr] = IntervalSet.from_tuples_with_id(tups)

    def has_overlapping(self, chr, start, stop):
        """is there an interval overlapping the region passed"""
        self.do_build_intervals()
        if not chr in self._interval_sets:
            return False
        return self._interval_sets[chr].has_overlap(start, stop)

    def get_overlapping(self, chr, start, stop):
        """Retrieve the rows for the region passed in.
        Returns an empty dataframe if there is no overlap

        Please note that the interval is end excluding - ie start == stop
        means an empty interval and nothing ever overlapping!
        """
        # print 'testing overlap for', chr, start, stop
        self.do_build_intervals()
        if not chr in self._interval_sets:
            return self.df[0:0]
        ids = sorted(self._interval_sets[chr].get_overlap(start, stop).to_ids())
        return self.df.loc[ids]

    def get_closest_by_start(self, chr, point):
        """Find the interval that has the  closest to the passed point.
        Returns a df with that interval, or an empty df!
        """
        # first we check whether we're on top of a region...
        # overlapping = self.get_overlapping(chr, point, point+1)
        # if len(overlapping):
        # return overlapping # could be more than one?!
        self.do_build_intervals()
        if not chr in self._interval_sets:
            return self.df[0:0]
        res = self._interval_sets[chr].find_closest_start(point)
        if res is None:  # pragma: no cover
            return self.df[0:0]
        start, stop, ids = res
        return self.df.loc[ids]

    # various statistics
    def get_no_of_entries(self):
        """How many intervals are there"""
        return len(self.df)

    @lazy_property
    def covered_bases(self):
        """How many base pairs are covered by these intervals"""
        if self.on_overlap == "ignore":
            raise ValueError(
                "covered_bases is currently only implemented for not overlapping GenomicRegions (on_overlap != 'ignore')"
            )
        return (self.df["stop"] - self.df["start"]).sum()

    @lazy_property
    def mean_size(self):
        """Get the mean size of the intervals defined in this GR"""
        return np.mean(self.df["stop"] - self.df["start"])

    def register(self):
        region_registry[self.name] = self

    # magic
    def __hash__(self):
        return hash("GR" + self.name)

    def __str__(self):
        return "GenomicRegion(%s)" % self.name

    def __repr__(self):
        return "GenomicRegion(%s)" % self.name

    # output functions

    def write_bed(self, output_filename=None, region_name=None, include_header=False):
        """Store the intervals of the GenomicRegion in a BED file
        @region_name: Insert the column of the GenomicRegions-Object e.g. 'repeat_name'"""
        from mbf_fileformats.bed import BedEntry, write_bed

        output_filename = self.pathify(output_filename, self.name + ".bed")

        def write(output_filename=output_filename):
            bed_entries = []
            for ii, row in self.df.iterrows():
                if region_name is None:

                    def get_name(row):
                        return None

                else:

                    def get_name(row):
                        if region_name in row:
                            return row[region_name]
                        else:
                            raise KeyError(
                                "key: %s not in genomic regions object. These objects are available: %s"
                                % (region_name, self.df.columns)
                            )

                entry = BedEntry(
                    row["chr"],
                    row["start"],
                    row["stop"],
                    name=get_name(row),
                    strand=row["strand"] if "strand" in row else None,
                    score=row["score"] if "score" in row else None,
                )
                bed_entries.append(entry)
            write_bed(
                output_filename,
                bed_entries,
                {},
                self.name,
                include_header=include_header,
            )

        if self.load_strategy.build_deps:
            deps = [
                self.load(),
                ppg.ParameterInvariant(output_filename, (include_header,)),
            ]
        else:
            deps = []
        return self.load_strategy.generate_file(output_filename, write, deps)

    def write_bigbed(
        self, output_filename=None, name_column=None
    ):  # pragma: no cover - till we have the test
        """Store the intervals of the GenomicRegion in a big bed file"""
        from mbf_fileformats.bed import BedEntry, write_bigbed

        output_filename = self.pathify(output_filename, self.name + ".bigbed")

        def write(output_filename=output_filename):

            bed_entries = []
            for idx, row in self.df.iterrows():
                entry = BedEntry(
                    row["chr"],
                    row["start"],
                    row["stop"],
                    name=row[name_column] if name_column else None,
                    strand=row["strand"] if "strand" in row else 0,
                )
                bed_entries.append(entry)
            if len(self.df):
                write_bigbed(
                    bed_entries, output_filename, self.genome.get_chromosome_lengths()
                )
            else:
                with open(output_filename, "wb"):
                    pass

        if self.load_strategy.build_deps:
            deps = [self.load()]
        else:
            deps = []
        return self.load_strategy.generate_file(
            output_filename, write, deps, empty_ok=True
        )

    # filtering
    def _new_for_filtering(self, new_name, load_func, deps, **kwargs):
        """When filtering, a new object of this class is created.
        To pass it the right options from the parent, overwrite this
        """
        kwargs["sheet_name"] = kwargs.get("sheet_name", "Filtered")
        for k in "on_overlap", "result_dir", "summit_annotator", "vid":
            if not k in kwargs:
                kwargs[k] = getattr(self, k)
        return GenomicRegions(new_name, load_func, deps, genome=self.genome, **kwargs)

    def _iter_intersections(self, other_gr):
        """Iterate over (chr, start, stop) tuples of the intersections between this GenomicRegions
        and other_gr.

        Refactored from intersection and overlap_basepairs"""
        if self.genome != other_gr.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_gr.genome)
            )
        if self.need_to_handle_overlapping_regions:
            raise ValueError(
                "_iter_intersections currently only works on non_overlapping intervalsets - %s was %s"
                % (self.name, self.on_overlap)
            )
        if other_gr.need_to_handle_overlapping_regions:
            other_df = merge_df_intervals(other_gr.df)
        else:
            other_df = other_gr.df
        for chr in self.df["chr"].unique():
            df_here = self.df[self.df["chr"] == chr]
            starts_here = df_here["start"]
            stops_here = df_here["stop"]
            print(other_df)
            print(other_df.dtypes)
            df_there = other_df[other_df["chr"] == chr]
            starts_there = df_there["start"]
            stops_there = df_there["stop"]
            ii = 0
            jj = 0
            while ii < len(df_here) and jj < len(df_there):
                if starts_there.iloc[jj] <= starts_here.iloc[ii] < stops_there.iloc[jj]:
                    yield (
                        chr,
                        max(starts_here.iloc[ii], starts_there.iloc[jj]),
                        min(stops_here.iloc[ii], stops_there.iloc[jj]),
                    )
                    ii += 1
                elif (
                    starts_here.iloc[ii] <= starts_there.iloc[jj] <= stops_here.iloc[ii]
                ):
                    yield (
                        chr,
                        max(starts_here.iloc[ii], starts_there.iloc[jj]),
                        min(stops_here.iloc[ii], stops_there.iloc[jj]),
                    )
                    jj += 1
                elif starts_here.iloc[ii] < starts_there.iloc[jj]:
                    ii += 1
                else:
                    jj += 1
        return

    def copy_annotators(self, *other_grs):
        """Copy annotators from other GRs"""
        for p in other_grs:
            for anno in p.annotators.values():
                self += anno

    # overlap calculations
    def overlap_basepair(
        self, other_gr
    ):  # todo: given to GRs with on_overlap != ignore, we could do a sorted search like we did for the GIS, that should be faster...
        """calculate the overlap between to GenomicRegions, on a base pair level - ie. the size of the intersection set
        (this converts other_gr into a disjoint set, so each base is counted only once. self must be disjoint)"""
        if self.genome != other_gr.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_gr.genome)
            )
        overlap = 0
        for chr, start, stop in self._iter_intersections(other_gr):
            overlap += stop - start
        return overlap

    def overlap_percentage(self, other_gr):
        """calculate the percentage of overlapping base pairs
        Basically overlap in bp / min(bp self, bp other_gr)
        if either is empty, percentage = 0
        """
        if self.genome != other_gr.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_gr.genome)
            )
        bp_overlap = self.overlap_basepair(other_gr)
        divisor = min(self.covered_bases, other_gr.covered_bases)
        if divisor == 0:
            return 0
        else:
            return float(bp_overlap) / divisor

    def overlap_count(self, other_gr):
        """Count the size of the union of all intersecting intervals. Basically, len(overlapping(..., other_gr))

        See L{overlapping}.

        Requires self.load() to have been done.

        This is (much) faster than building overlap()  and calling len,.

        The invariant len(a) + len(b) == len(a.filter_remove_overlapping(b))
            + len(b.filter_remove_overlapping(a)) + a.overlap_count(b) is not true,
             because the overlapping count only counts each region once,
             even if it was compromised out of multiple regions in the input. Say you have
        ---  ----  -----
          ----  ----
        that count's as 3 (or two) regions in len(a) (len(b)),
         but just as one in overlapping/overlap_count (and the filtered ones are of len(0)). Still, I argue that this is the right one to use in venn diagrams.

        """
        if self.genome != other_gr.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_gr.genome)
            )
        overlap = 0
        for chrom in self.genome.get_chromosome_lengths():
            a = self.df[self.df.chr == chrom][["start", "stop"]].assign(color=1)
            b = other_gr.df[other_gr.df.chr == chrom][["start", "stop"]].assign(color=2)
            comb = pd.concat([a, b]).sort_values("start")
            last_stop = 0
            run_start = -1
            run_colors = 0
            for dummy_idx, row in comb.iterrows():
                start = row["start"]
                stop = row["stop"]
                color = row["color"]
                if start < last_stop:  # a continuation
                    run_colors |= color
                    last_stop = max(last_stop, stop)
                else:  # end of this run
                    if run_start > 0:
                        if run_colors == 3:
                            overlap += 1
                    run_start = start
                    last_stop = stop
                    run_colors = color
            if (
                len(comb) and start != run_start and run_colors == 3
            ):  # capture the last one that will not have triggered an 'end-of-run'
                overlap += 1
        return overlap

    def intersection_count(self, other_gr):
        """calculate the number of intersecting regions between two GenomicRegions
        Note that
        len(here) - intersection_count(here, other_gr) != len(here.filter_remove_overlapping(...,other_gr)
        since they handle this situation differently:
        here : ---    ----
        there:   ------
        (which here & in intersection gives you two intervals, while in overlapping (and overlap_count) it
        gives just one)

        """
        if self.genome != other_gr.genome:
            raise ValueError(
                "GenomicRegions set-operations only work if both have the same genome. You had %s and %s"
                % (self.genome, other_gr.genome)
            )
        if not self.load_strategy.build_deps:
            other_gr.load()
        overlap = 0
        for chr, start, stop in self._iter_intersections(other_gr):
            overlap += 1
        return overlap

    def convert(
        self,
        new_name,
        conversion_function,
        new_genome=None,
        dependencies=None,
        on_overlap="raise",
        sheet_name="Converted",
        anno_columns_to_keep=[],
        summit_annotator=None,
    ):
        """Convert the intervals into some other coordinate system,
        possibly changing the genome as well.
        Does not conserve annotators, non canonical rows depend on the conversion fucntion

        @conversion_function is passed the dataframe, and must return one containing
        at least (chr, start, stop).
        @conversion_function may be a tuple (function, [annotators]),
            in which case the function is treated as conversion_function
            and the annotators are added as dependencies
        @conversion_function may be a tuple (function, [annotators], parameters),
            in which case the function is treated as conversion_function and the
            annotators are added as dependencies,
            and an additional parameter_dependency is added
        @dependencies must be a list of jobs
        """
        if not isinstance(new_name, str):
            raise ValueError("Name must be a string")
        if isinstance(conversion_function, tuple):
            if len(conversion_function) > 2:
                convert_parameters = conversion_function[2]
            else:
                convert_parameters = None
            annotators_required = conversion_function[1]
            conversion_function = conversion_function[0]
        else:
            annotators_required = []
            convert_parameters = None

        def do_load():
            df = conversion_function(self.df)
            if not isinstance(df, pd.DataFrame):
                raise ValueError(
                    "GenomicRegions.convert conversion_function must return a pandas.DataFrame."
                )
            for col in df.columns[:]:
                if not (
                    col in self.non_annotator_columns or col in anno_columns_to_keep
                ):
                    df = df.drop(col, axis=1)
            return df

        if new_genome is None:
            new_genome = self.genome
        else:
            if not isinstance(new_genome, GenomeBase):
                raise ValueError(
                    "new_genome %s was not a genome. Did you mean to pass in dependencies, that's the next parameter?"
                    % new_genome
                )
        if self.load_strategy.build_deps:
            deps = [self.load()]
            if dependencies:
                deps.extend(dependencies)
            for anno in annotators_required:
                deps.append(self.add_annotator(anno))
            deps.append(
                ppg.ParameterInvariant(
                    new_name + "_conversion_paramaters", convert_parameters
                )
            )
            deps.append(
                ppg.FunctionInvariant(
                    new_name + "_conversion_function", conversion_function
                )
            )
        else:
            deps = []
        if hasattr(conversion_function, "dependencies"):
            deps.extend(conversion_function.dependencies)
        if summit_annotator is None:  # pragma: no cover
            summit_annotator = self.summit_annotator
        return GenomicRegions(
            new_name,
            do_load,
            deps,
            new_genome,
            on_overlap=on_overlap,
            summit_annotator=summit_annotator,
            sheet_name=sheet_name,
            vid=self.vid,
        )



from .delayeddataframe import DelayedDataFrame  # noqa:F401
from . import annotator  # noqa: F401
from . import regions
from . import genes

__version__ = '0.3'


_all__ = [DelayedDataFrame, annotator, regions, genes, __version__]



from abc import ABC
from typing import List
import pandas as pd
import pypipegraph as ppg
from .util import freeze
from pathlib import Path
import numpy as np

annotator_singletons = {"lookup": []}


class Annotator(ABC):
    def __new__(cls, *args, **kwargs):
        cn = cls.__name__
        if ppg.inside_ppg():
            if not hasattr(ppg.util.global_pipegraph, "_annotator_singleton_dict"):
                ppg.util.global_pipegraph._annotator_singleton_dict = {"lookup": []}
            singleton_dict = ppg.util.global_pipegraph._annotator_singleton_dict
        else:
            singleton_dict = annotator_singletons
        if not cn in singleton_dict:
            singleton_dict[cn] = {}
        key = {}
        for ii in range(0, len(args)):
            key["arg_%i" % ii] = args[ii]
        key.update(kwargs)
        for k, v in key.items():
            key[k] = freeze(v)
        key = tuple(sorted(key.items()))
        if not key in singleton_dict[cn]:
            singleton_dict[cn][key] = object.__new__(cls)
            singleton_dict["lookup"].append(singleton_dict[cn][key])

        return singleton_dict[cn][key]

    def __hash__(self):
        return hash(self.get_cache_name())

    def __str__(self):
        return "Annotator %s" % self.columns[0]

    def __repr__(self):
        return "Annotator(%s)" % self.columns[0]

    def get_cache_name(self):
        if hasattr(self, "cache_name"):
            return self.cache_name
        else:
            return self.columns[0]

    def calc(self, df):
        raise NotImplementedError()  # pragma: no cover

    def deps(self, ddf):
        """Return ppg.jobs"""
        return []

    def dep_annos(self):
        """Return other annotators"""
        return []


class Constant(Annotator):
    def __init__(self, column_name, value):
        self.columns = [column_name]
        self.value = value

    def calc(self, df):
        return pd.DataFrame({self.columns[0]: self.value}, index=df.index)


class FromFile(Annotator):
    def __init__(
        self,
        tablepath: Path,
        columns_to_add: List[str],
        index_column_table: str = "gene_stable_id",
        index_column_genes: str = "gene_stable_id",
        fill_value: float = None,
    ):
        """
        Adds arbitrary columns from a table.

        This requires that both the table and the ddf have a common column on
        which we can index.

        Parameters
        ----------
        tablepath : Path
            Path to table with additional columns.
        columns_to_add : List[str]
            List of columns to append.
        index_column_table : str, optional
            Index column in table, by default "gene_stable_id".
        index_column_genes : str, optional
            Index column in ddf to append to, by default "gene_stable_id".
        fill_value : float, optonal
            Value to fill for missing rows, defaults to np.NaN.
        """
        self.tablepath = tablepath
        self.columns = columns_to_add
        self.index_column_table = index_column_table
        self.index_column_genes = index_column_genes
        self.fill = fill_value if fill_value is not None else np.NaN

    def parse(self):
        if (self.tablepath.suffix == ".xls") or (self.tablepath.suffix == ".xlsx"):
            return pd.read_excel(self.tablepath)
        else:
            return pd.read_csv(self.tablepath, sep="\t")

    def get_cache_name(self):
        return f"FromFile_{self.tablepath.name}"

    def calc_ddf(self, ddf):
        """Calculates the ddf to append."""
        df_copy = ddf.df.copy()
        if self.index_column_genes not in df_copy.columns:
            raise ValueError(
                f"Column {self.index_column_genes} not found in ddf index, found was:\n{[str(x) for x in df_copy.columns]}."
            )
        df_in = self.parse()
        if self.index_column_table not in df_in.columns:
            raise ValueError(
                f"Column {self.index_column_table} not found in table, found was:\n{[str(x) for x in df_in.columns]}."
            )
        for column in self.columns:
            if column not in df_in.columns:
                raise ValueError(
                    f"Column {column} not found in table, found was:\n{[str(x) for x in df_in.columns]}."
                )
        df_copy.index = df_copy[self.index_column_genes]
        df_in.index = df_in[self.index_column_table]
        df_in = df_in.reindex(df_copy.index, fill_value=self.fill)
        df_in = df_in[self.columns]
        df_in.index = ddf.df.index
        return df_in

    def deps(self, ddf):
        """Return ppg.jobs"""
        return ppg.FileInvariant(self.tablepath)



"""New style (fast) tag count annos

Use these for new projects.

"""
from mbf_genomics.annotator import Annotator
import numpy as np
import pypipegraph as ppg
import hashlib
import pandas as pd
from pathlib import Path
from dppd import dppd
import dppd_plotnine  # noqa:F401
from mbf_qualitycontrol import register_qc, QCCollectingJob, qc_disabled
from mbf_genomics.util import parse_a_or_c_to_plot_name

dp, X = dppd()


# ## Base classes and strategies - skip these if you just care about using TagCount annotators
class _CounterStrategyBase:
    cores_needed = 1

    def extract_lookup(self, data):
        """Adapter for count strategies that have different outputs
        (e.g. one-hashmap-unstranded or two-hashmaps-one-forward-one-reversed)
        """
        return data


class CounterStrategyStrandedRust(_CounterStrategyBase):
    cores_needed = -1
    name = "stranded"

    def __init__(self):
        self.disable_sanity_check = False

    def count_reads(
        self, interval_strategy, genome, bam_filename, bam_index_name, reverse=False
    ):
        # bam_filename = bamfil

        intervals = interval_strategy._get_interval_tuples_by_chr(genome)
        gene_intervals = IntervalStrategyGene()._get_interval_tuples_by_chr(genome)
        from mbf_bam import count_reads_stranded

        res = count_reads_stranded(
            bam_filename, bam_index_name, intervals, gene_intervals
        )
        self.sanity_check(res, bam_filename)
        return res

    def sanity_check(self, foward_and_reverse, bam_filename):
        if self.disable_sanity_check:
            return
        error_count = 0
        forward, reverse = foward_and_reverse
        for gene_stable_id, forward_count in forward.items():
            reverse_count = reverse.get(gene_stable_id, 0)
            if (reverse_count > 100) and (reverse_count > forward_count * 1.1):
                error_count += 1
        if error_count > 0.1 * len(forward):
            raise ValueError(
                "Found at least %.2f%% of genes to have a reverse read count (%s) "
                "above 110%% of the exon read count (and at least 100 tags). "
                "This indicates that this lane (%s) should have been reversed before alignment. "
                "Set reverse_reads=True on your Lane object"
                % (
                    100.0 * error_count / len(forward),
                    self.__class__.__name__,
                    bam_filename,
                )
            )

    def extract_lookup(self, data):
        """Adapter for count strategies that have different outputs
        (e.g. one-hashmap-unstranded or two-hashmaps-one-forward-one-reversed)
        """
        return data[0]


class CounterStrategyUnstrandedRust(_CounterStrategyBase):
    cores_needed = -1
    name = "unstranded"

    def count_reads(
        self, interval_strategy, genome, bam_filename, bam_index_name, reverse=False
    ):
        # bam_filename = bamfil

        intervals = interval_strategy._get_interval_tuples_by_chr(genome)
        gene_intervals = IntervalStrategyGene()._get_interval_tuples_by_chr(genome)
        # chr -> [gene_id, strand, [start], [stops]
        from mbf_bam import count_reads_unstranded

        res = count_reads_unstranded(
            bam_filename, bam_index_name, intervals, gene_intervals
        )
        return res


class _IntervalStrategy:
    def get_interval_lengths_by_gene(self, genome):
        by_chr = self._get_interval_tuples_by_chr(genome)
        length_by_gene = {}
        for chr, tups in by_chr.items():
            for tup in tups:  # stable_id, strand, [starts], [stops]
                gene_stable_id = tup[0]
                length = 0
                for start, stop in zip(tup[2], tup[3]):
                    length += stop - start
                length_by_gene[gene_stable_id] = length
        return length_by_gene

    def _get_interval_tuples_by_chr(self, genome):  # pragma: no cover
        raise NotImplementedError()

    def get_deps(self):
        return []


class IntervalStrategyGenomicRegion(_IntervalStrategy):
    """Used internally by _FastTagCounterGR"""

    def __init__(self, gr):
        self.gr = gr
        self.name = f"GR_{gr.name}"

    def _get_interval_tuples_by_chr(self, genome):
        result = {chr: [] for chr in genome.get_chromosome_lengths()}
        if self.gr.genome != genome:  # pragma: no cover
            raise ValueError("Mismatched genomes")
        df = self.gr.df
        if not "strand" in df.columns:
            df = df.assign(strand=1)
        df = df[["chr", "start", "stop", "strand"]]
        if df.index.duplicated().any():
            raise ValueError("index must be unique")
        for tup in df.itertuples():
            result[tup.chr].append((str(tup[0]), tup.strand, [tup.start], [tup.stop]))
        return result


class IntervalStrategyGene(_IntervalStrategy):
    """Count from TSS to TES"""

    name = "gene"

    def _get_interval_tuples_by_chr(self, genome):
        result = {chr: [] for chr in genome.get_chromosome_lengths()}
        gene_info = genome.df_genes
        for tup in gene_info[["chr", "start", "stop", "strand"]].itertuples():
            result[tup.chr].append((tup[0], tup.strand, [tup.start], [tup.stop]))
        return result


class IntervalStrategyExon(_IntervalStrategy):
    """count all exons"""

    name = "exon"

    def _get_interval_tuples_by_chr(self, genome):
        result = {chr: [] for chr in genome.get_chromosome_lengths()}
        for gene in genome.genes.values():
            exons = gene.exons_merged
            result[gene.chr].append(
                (gene.gene_stable_id, gene.strand, list(exons[0]), list(exons[1]))
            )
        return result


class IntervalStrategyIntron(_IntervalStrategy):
    """count all introns"""

    name = "intron"

    def _get_interval_tuples_by_chr(self, genome):
        result = {chr: [] for chr in genome.get_chromosome_lengths()}
        for gene in genome.genes.values():
            exons = gene.introns_strict
            result[gene.chr].append(
                (gene.gene_stable_id, gene.strand, list(exons[0]), list(exons[1]))
            )
        return result


class IntervalStrategyExonSmart(_IntervalStrategy):
    """For protein coding genes: count only in exons of protein-coding transcripts.
    For other genes: count all exons"""

    name = "exonsmart"

    def _get_interval_tuples_by_chr(self, genome):
        result = {chr: [] for chr in genome.get_chromosome_lengths()}
        for g in genome.genes.values():
            e = g.exons_protein_coding_merged
            if len(e[0]) == 0:
                e = g.exons_merged
            result[g.chr].append((g.gene_stable_id, g.strand, list(e[0]), list(e[1])))
        return result


# Now the actual tag count annotators
class TagCountCommonQC:
    def register_qc(self, genes):
        if not qc_disabled():
            self.register_qc_distribution(genes)
            self.register_qc_pca(genes)
            # self.register_qc_cummulative(genes)

    def register_qc_distribution(self, genes):
        output_filename = genes.result_dir / self.qc_folder / "read_distribution.png"
        output_filename.parent.mkdir(exist_ok=True)

        def plot(output_filename, elements):
            df = genes.df
            df = dp(df).select({x.aligned_lane.name: x.columns[0] for x in elements}).pd
            if len(df) == 0:
                df = pd.DataFrame({"x": [0], "y": [0], "text": "no data"})
                dp(df).p9().add_text("x", "y", "text").render(output_filename).pd
            else:
                plot_df = dp(df).melt(var_name="sample", value_name="count").pd

                plot = dp(plot_df).p9().theme_bw()
                print(df)

                if ((df > 0).sum(axis=0) > 1).any() and len(df) > 1:
                    plot = plot.geom_violin(
                        dp.aes(x="sample", y="count"), width=0.5, bw=0.1
                    )
                if len(plot_df["sample"].unique()) > 1:
                    plot = plot.annotation_stripes(fill_range=True)
                if (plot_df["count"] > 0).any():
                    # can't have a log boxplot with all nans (log(0))
                    plot = plot.scale_y_continuous(
                        trans="log10",
                        name=self.qc_distribution_scale_y_name,
                        breaks=[1, 10, 100, 1000, 10000, 100_000, 1e6, 1e7],
                    )
                print(plot_df)

                return (
                    plot.add_boxplot(
                        x="sample", y="count", _width=0.1, _fill=None, _color="blue"
                    )
                    .turn_x_axis_labels()
                    .title("Raw read distribution")
                    .hide_x_axis_title()
                    .render_args(limitsize=False)
                    .render(output_filename, width=0.2 * len(elements) + 1, height=4)
                )

        return register_qc(
            QCCollectingJob(output_filename, plot)
            .depends_on(genes.add_annotator(self))
            .add(self)
        )

    def register_qc_pca(self, genes):
        output_filename = genes.result_dir / self.qc_folder / "pca.png"

        def plot(output_filename, elements):
            import sklearn.decomposition as decom

            if len(elements) == 1:
                xy = np.array([[0], [0]]).transpose()
                title = "PCA %s - fake / single sample" % genes.name
            else:
                pca = decom.PCA(n_components=2, whiten=False)
                data = genes.df[[x.columns[0] for x in elements]]
                data -= data.min()  # min max scaling 0..1
                data /= data.max()
                data = data[~pd.isnull(data).any(axis=1)]  # can' do pca on NAN values
                if len(data):
                    pca.fit(data.T)
                    xy = pca.transform(data.T)
                    title = "PCA %s\nExplained variance: x %.2f%%, y %.2f%%" % (
                        genes.name,
                        pca.explained_variance_ratio_[0] * 100,
                        pca.explained_variance_ratio_[1] * 100,
                    )
                else:
                    xy = np.array(
                        [[0] * len(elements), [0] * len(elements)]
                    ).transpose()
                    title = "PCA %s - fake / no rows" % genes.name

            plot_df = pd.DataFrame(
                {"x": xy[:, 0], "y": xy[:, 1], "label": [x.plot_name for x in elements]}
            )
            print(plot_df)
            (
                dp(plot_df)
                .p9()
                .theme_bw()
                .add_scatter("x", "y")
                .add_text(
                    "x",
                    "y",
                    "label",
                    # cool, this can go into an endless loop...
                    # _adjust_text={
                    # "expand_points": (2, 2),
                    # "arrowprops": {"arrowstyle": "->", "color": "red"},
                    # },
                )
                .scale_color_many_categories()
                .title(title)
                .render(output_filename, width=8, height=6)
            )

        return register_qc(
            QCCollectingJob(output_filename, plot)
            .depends_on(genes.add_annotator(self))
            .add(self)
        )


class _FastTagCounter(Annotator, TagCountCommonQC):
    def __init__(
        self, aligned_lane, count_strategy, interval_strategy, column_name, column_desc
    ):
        if not hasattr(aligned_lane, "get_bam"):
            raise ValueError("_FastTagCounter only accepts aligned lanes!")
        self.aligned_lane = aligned_lane
        self.genome = self.aligned_lane.genome
        self.count_strategy = count_strategy
        self.interval_strategy = interval_strategy
        self.columns = [(column_name % (self.aligned_lane.name,)).strip()]
        self.cache_name = (
            "FT_%s_%s" % (count_strategy.name, interval_strategy.name)
            + "_"
            + hashlib.md5(self.columns[0].encode("utf-8")).hexdigest()
        )
        self.column_properties = {self.columns[0]: {"description": column_desc}}
        self.vid = aligned_lane.vid
        self.cores_needed = count_strategy.cores_needed
        self.plot_name = self.aligned_lane.name
        self.qc_folder = f"{self.count_strategy.name}_{self.interval_strategy.name}"
        self.qc_distribution_scale_y_name = "raw counts"

    def calc(self, df):
        if ppg.inside_ppg():
            data = self._data
        else:
            data = self.calc_data()
        lookup = self.count_strategy.extract_lookup(data)
        result = []
        for gene_stable_id in df["gene_stable_id"]:
            result.append(lookup.get(gene_stable_id, 0))
        result = np.array(result, dtype=np.float)
        return pd.Series(result)

    def deps(self, _genes):
        return [self.load_data()]

    def calc_data(self):
        bam_file, bam_index_name = self.aligned_lane.get_bam_names()
        return self.count_strategy.count_reads(
            self.interval_strategy, self.genome, bam_file, bam_index_name
        )

    def load_data(self):
        cf = Path(ppg.util.global_pipegraph.cache_folder) / "FastTagCounters"
        cf.mkdir(exist_ok=True)
        return (
            ppg.CachedAttributeLoadingJob(
                cf / self.cache_name, self, "_data", self.calc_data
            )
            .depends_on(self.aligned_lane.load())
            .use_cores(-1)
        )


class _FastTagCounterGR(Annotator):
    def __init__(self, aligned_lane, count_strategy, column_name, column_desc):
        if not hasattr(aligned_lane, "get_bam"):
            raise ValueError("_FastTagCounter only accepts aligned lanes!")
        self.aligned_lane = aligned_lane
        self.genome = self.aligned_lane.genome
        self.count_strategy = count_strategy
        self.columns = [(column_name % (self.aligned_lane.name,)).strip()]
        self.cache_name = (
            "FT_%s_%s" % (count_strategy.name, "on_gr")
            + "_"
            + hashlib.md5(self.columns[0].encode("utf-8")).hexdigest()
        )
        self.column_properties = {self.columns[0]: {"description": column_desc}}
        self.vid = aligned_lane.vid
        self.cores_needed = count_strategy.cores_needed
        self.plot_name = self.aligned_lane.name
        # self.qc_folder = f"{self.count_strategy.name}_{self.interval_strategy.name}"
        # self.qc_distribution_scale_y_name = "raw counts"

    def calc(self, df):
        if ppg.inside_ppg():
            data = self._data
        else:
            data = self.calc_data()
        lookup = self.count_strategy.extract_lookup(data)
        result = []
        for idx in df.index:
            result.append(lookup.get(str(idx), 0))
        result = np.array(result, dtype=np.float)
        return pd.Series(result)

    def deps(self, gr):
        return [self.load_data(gr)]

    def calc_data(self, gr):
        def inner():
            bam_file, bam_index_name = self.aligned_lane.get_bam_names()
            return self.count_strategy.count_reads(
                IntervalStrategyGenomicRegion(gr), self.genome, bam_file, bam_index_name
            )

        return inner

    def load_data(self, gr):
        cf = gr.cache_dir
        cf.mkdir(exist_ok=True)
        return (
            ppg.CachedAttributeLoadingJob(
                cf / self.cache_name, self, "_data", self.calc_data(gr)
            )
            .depends_on(self.aligned_lane.load())
            .depends_on(gr.load())
            .use_cores(-1)  # should be count_strategy cores needed, no?
        )


#
# ## Raw tag count annos for analysis usage


class ExonSmartStrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyStrandedRust(),
            IntervalStrategyExonSmart(),
            "Exon, protein coding, stranded smart tag count %s",
            "Tag count inside exons of protein coding transcripts (all if no protein coding transcripts) exons, correct strand only",
        )


class ExonSmartUnstrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyUnstrandedRust(),
            IntervalStrategyExonSmart(),
            "Exon, protein coding, unstranded smart tag count %s",
            "Tag count inside exons of protein coding transcripts (all if no protein coding transcripts)  both strands",
        )


class ExonStrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyStrandedRust(),
            IntervalStrategyExon(),
            "Exon, protein coding, stranded tag count %s",
            "Tag count inside exons of protein coding transcripts (all if no protein coding transcripts) exons, correct strand only",
        )


class ExonUnstrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyUnstrandedRust(),
            IntervalStrategyExon(),
            "Exon, protein coding, unstranded tag count %s",
            "Tag count inside exons of protein coding transcripts (all if no protein coding transcripts)  both strands",
        )


class GeneStrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyStrandedRust(),
            IntervalStrategyGene(),
            "Gene, stranded tag count %s",
            "Tag count inside gene body (tss..tes), correct strand only",
        )


class GeneUnstrandedRust(_FastTagCounter):
    def __init__(self, aligned_lane):
        _FastTagCounter.__init__(
            self,
            aligned_lane,
            CounterStrategyUnstrandedRust(),
            IntervalStrategyGene(),
            "Gene unstranded tag count %s",
            "Tag count inside gene body (tss..tes), both strands",
        )


def GRUnstrandedRust(aligned_lane):
    return _FastTagCounterGR(
        aligned_lane,
        CounterStrategyUnstrandedRust(),
        "Tag count %s",
        "Tag count inside region, both strands",
    )


def GRStrandedRust(aligned_lane):
    return _FastTagCounterGR(
        aligned_lane,
        CounterStrategyStrandedRust(),
        "Tag count %s",
        "Tag count inside region, stranded",
    )


# we are keeping the python ones for now as reference implementations
GeneUnstranded = GeneUnstrandedRust
GeneStranded = GeneStrandedRust
ExonStranded = ExonStrandedRust
ExonUnstranded = ExonUnstrandedRust
ExonSmartStranded = ExonSmartStrandedRust
ExonSmartUnstranded = ExonSmartUnstrandedRust

# ## Normalizing annotators - convert raw tag counts into something normalized


class _NormalizationAnno(Annotator, TagCountCommonQC):
    def __init__(self, base_column_spec):
        from ..util import parse_a_or_c_to_anno, parse_a_or_c_to_column

        self.raw_anno = parse_a_or_c_to_anno(base_column_spec)
        self.raw_column = parse_a_or_c_to_column(base_column_spec)
        if self.raw_anno is not None:
            self.genome = self.raw_anno.genome
            self.vid = getattr(self.raw_anno, "vid", None)
            self.aligned_lane = getattr(self.raw_anno, "aligned_lane", None)
        else:
            self.genome = None
            self.vid = None
            self.aligned_lane = None
        self.columns = [self.raw_column + " " + self.name]
        self.cache_name = (
            self.__class__.__name__
            + "_"
            + hashlib.md5(self.columns[0].encode("utf-8")).hexdigest()
        )
        if self.raw_anno is not None:
            self.plot_name = getattr(self.raw_anno, "plot_name", self.raw_column)
            if hasattr(self.raw_anno, "count_strategy"):
                if hasattr(self.raw_anno, "interval_strategy"):
                    iv_name = self.raw_anno.interval_strategy.name
                else:
                    iv_name = "-"
                self.qc_folder = f"normalized_{self.name}_{self.raw_anno.count_strategy.name}_{iv_name}"
            else:
                self.qc_folder = f"normalized_{self.name}"
        else:
            self.plot_name = parse_a_or_c_to_plot_name(base_column_spec)
            self.qc_folder = f"normalized_{self.name}"
        self.qc_distribution_scale_y_name = self.name

    def dep_annos(self):
        if self.raw_anno is None:
            return []
        else:
            return [self.raw_anno]


class NormalizationCPM(_NormalizationAnno):
    """Normalize to 1e6 by taking the sum of all genes"""

    def __init__(self, base_column_spec):
        self.name = "CPM"
        self.normalize_to = 1e6
        super().__init__(base_column_spec)
        self.column_properties = {
            self.columns[0]: {
                "description": "Tag count inside protein coding (all if no protein coding transcripts) exons, normalized to 1e6 across all genes"
            }
        }

    def calc(self, df):
        raw_counts = df[self.raw_column]
        total = max(1, float(raw_counts.sum()))  # avoid division by 0
        result = raw_counts * (self.normalize_to / total)
        return pd.Series(result)


class NormalizationTPM(_NormalizationAnno):
    """Normalize to transcripts per million, ie.
        count / length * (1e6 / (sum_i(count_/length_i)))

    """

    def __init__(self, base_column_spec, interval_strategy=None):
        self.name = "TPM"
        self.normalize_to = 1e6
        super().__init__(base_column_spec)
        if self.raw_anno is None:  # pragma: no cover
            if interval_strategy is None:  # pragma: no cover
                raise ValueError(
                    "TPM normalization needs to know the intervals used. Either base of a FastTagCount annotator or pass in an interval strategy"
                )
            self.interval_strategy = interval_strategy
        else:
            self.interval_strategy = self.raw_anno.interval_strategy
        self.column_properties = {
            self.columns[0]: {"description": "transcripts per million"}
        }

    def calc(self, df):
        raw_counts = df[self.raw_column]
        length_by_gene = self.interval_strategy.get_interval_lengths_by_gene(
            self.genome
        )
        result = np.zeros(raw_counts.shape, float)
        for ii, gene_stable_id in enumerate(df["gene_stable_id"]):
            result[ii] = raw_counts.iloc[ii] / length_by_gene[gene_stable_id]
        total = float(result[~pd.isnull(result)].sum())
        factor = 1e6 / total
        result = result * factor
        return pd.DataFrame({self.columns[0]: result})


class NormalizationFPKM(Annotator):
    def __init__(self, raw_anno):
        raise NotImplementedError(
            "FPKM is a bad thing to use. It is not supported by mbf"
        )


class Salmon(Annotator):
    """Add salmon gene level estimation calculated on a raw Sample
    """

    def __init__(
        self,
        raw_lane,
        prefix="Salmon",
        options={
            # "--validateMappings": None,  this always get's set by aligners.Salmon
            "--gcBias": None,
            "--seqBias": None,
        },
        libtype="A",
        accepted_biotypes=None,  # set(("protein_coding", "lincRNA")),
        salmon_version="_last_used",
    ):
        self.raw_lane = raw_lane
        self.options = options.copy()
        self.libtype = libtype
        self.accepted_biotypes = accepted_biotypes
        self.salmon_version = salmon_version
        self.columns = [
            f"{prefix} TPM {raw_lane.name}",
            f"{prefix} NumReads {raw_lane.name}",
        ]

    def deps(self, ddf):
        import mbf_externals

        return mbf_externals.aligners.Salmon(
            self.accepted_biotypes, version=self.salmon_version
        ).run_quant_on_raw_lane(
            self.raw_lane, ddf.genome, self.libtype, self.options, gene_level=True
        )

    def calc_ddf(self, ddf):
        quant_path = Path(self.deps(ddf).job_id).parent / "quant.genes.sf"
        in_df = pd.read_csv(quant_path, sep="\t").set_index("Name")[["TPM", "NumReads"]]
        in_df.columns = self.columns
        res = in_df.reindex(ddf.df.gene_stable_id)
        res.index = ddf.df.index
        return res
